# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
import socket

from rekall import kb
from rekall import obj
from rekall import utils
from rekall.plugins.overlays.windows import pe_vtypes


AF_INET = 2
AF_INET6 = 0x17

# String representations of INADDR_ANY and INADDR6_ANY
inaddr_any = utils.inet_ntop(socket.AF_INET, '\0' * 4)
inaddr6_any = utils.inet_ntop(socket.AF_INET6, '\0' * 16)

protos = {
    0:"HOPOPT",
    1:"ICMP",
    2:"IGMP",
    3:"GGP",
    4:"IPv4",
    5:"ST",
    6:"TCP",
    7:"CBT",
    8:"EGP",
    9:"IGP",
    10:"BBN-RCC-MON",
    11:"NVP-II",
    12:"PUP",
    13:"ARGUS",
    14:"EMCON",
    15:"XNET",
    16:"CHAOS",
    17:"UDP",
    18:"MUX",
    19:"DCN-MEAS",
    20:"HMP",
    21:"PRM",
    22:"XNS-IDP",
    23:"TRUNK-1",
    24:"TRUNK-2",
    25:"LEAF-1",
    26:"LEAF-2",
    27:"RDP",
    28:"IRTP",
    29:"ISO-TP4",
    30:"NETBLT",
    31:"MFE-NSP",
    32:"MERIT-INP",
    33:"DCCP",
    34:"3PC",
    35:"IDPR",
    36:"XTP",
    37:"DDP",
    38:"IDPR-CMTP",
    39:"TP++",
    40:"IL",
    41:"IPv6",
    42:"SDRP",
    43:"IPv6-Route",
    44:"IPv6-Frag",
    45:"IDRP",
    46:"RSVP",
    47:"GRE",
    48:"DSR",
    49:"BNA",
    50:"ESP",
    51:"AH",
    52:"I-NLSP",
    53:"SWIPE",
    54:"NARP",
    55:"MOBILE",
    56:"TLSP",
    57:"SKIP",
    58:"IPv6-ICMP",
    59:"IPv6-NoNxt",
    60:"IPv6-Opts",
    61:"Host-interal",
    62:"CFTP",
    63:"Local Network",
    64:"SAT-EXPAK",
    65:"KRYPTOLAN",
    66:"RVD",
    67:"IPPC",
    68:"Dist-FS",
    69:"SAT-MON",
    70:"VISA",
    71:"IPCV",
    72:"CPNX",
    73:"CPHB",
    74:"WSN",
    75:"PVP",
    76:"BR-SAT-MON",
    77:"SUN-ND",
    78:"WB-MON",
    79:"WB-EXPAK",
    80:"ISO-IP",
    81:"VMTP",
    82:"SECURE-VMTP",
    83:"VINES",
    84:"TTP",
    # 84:"IPTM",
    85:"NSFNET-IGP",
    86:"DGP",
    87:"TCF",
    88:"EIGRP",
    89:"OSPFIGP",
    90:"Sprite-RPC",
    91:"LARP",
    92:"MTP",
    93:"AX.25",
    94:"IPIP",
    95:"MICP",
    96:"SCC-SP",
    97:"ETHERIP",
    98:"ENCAP",
    99:"Encryption",
    100:"GMTP",
    101:"IFMP",
    102:"PNNI",
    103:"PIM",
    104:"ARIS",
    105:"SCPS",
    106:"QNX",
    107:"A/N",
    108:"IPComp",
    109:"SNP",
    110:"Compaq-Peer",
    111:"IPX-in-IP",
    112:"VRRP",
    113:"PGM",
    114:"0-hop",
    115:"L2TP",
    116:"DDX",
    117:"IATP",
    118:"STP",
    119:"SRP",
    120:"UTI",
    121:"SMP",
    122:"SM",
    123:"PTP",
    124:"ISIS over IPv4",
    125:"FIRE",
    126:"CRTP",
    127:"CRUDP",
    128:"SSCOPMCE",
    129:"IPLT",
    130:"SPS",
    131:"PIPE",
    132:"SCTP",
    133:"FC",
    134:"RSVP-E2E-IGNORE",
    135:"Mobility Header",
    136:"UDPLite",
    137:"MPLS-in-IP",
    138:"manet",
    139:"HIP",
    140:"Shim6",
    141:"WESP",
    142:"ROHC",
    253:"Experimental",
    254:"Experimental",
    255:"Reserved",
}

# Structures used by connections, connscan, sockets, sockscan.
# Used by x86 XP (all service packs) and x86 2003 SP0.
tcpip_vtypes = {
    '_ADDRESS_OBJECT' : [0x68, {
        'Next' : [0x0, ['pointer', ['_ADDRESS_OBJECT']]],
        'LocalIpAddress' : [0x2c, ['Ipv4Address']],
        'LocalPort' : [0x30, ['unsigned be short']],
        'Protocol'  : [0x32, ['unsigned short']],
        'Pid' : [0x148, ['unsigned long']],
        'CreateTime' : [0x158, ['WinFileTime', {}]],
    }],
    '_TCPT_OBJECT' : [0x20, {
        'Next' : [0x0, ['pointer', ['_TCPT_OBJECT']]],
        'RemoteIpAddress' : [0xc, ['Ipv4Address']],
        'LocalIpAddress' : [0x10, ['Ipv4Address']],
        'RemotePort' : [0x14, ['unsigned be short']],
        'LocalPort' : [0x16, ['unsigned be short']],
        'Pid' : [0x18, ['unsigned long']],
    }],
}

# Structures used by connections, connscan, sockets, sockscan.
# Used by x64 XP and x64 2003 (all service packs).
tcpip_vtypes_2003_x64 = {
    '_ADDRESS_OBJECT' : [0x250, {
        'Next' : [0x0, ['pointer', ['_ADDRESS_OBJECT']]],
        'LocalIpAddress' : [0x58, ['Ipv4Address']],
        'LocalPort' : [0x5c, ['unsigned be short']],
        'Protocol'  : [0x5e, ['unsigned short']],
        'Pid' : [0x238, ['unsigned long']],
        'CreateTime' : [0x248, ['WinFileTime', {}]],
    }],
    '_TCPT_OBJECT' : [0x28, {
        'Next' : [0x0, ['pointer', ['_TCPT_OBJECT']]],
        'RemoteIpAddress' : [0x14, ['Ipv4Address']],
        'LocalIpAddress' : [0x18, ['Ipv4Address']],
        'RemotePort' : [0x1c, ['unsigned be short']],
        'LocalPort' : [0x1e, ['unsigned be short']],
        'Pid' : [0x20, ['unsigned long']],
    }],
}

# Structures used by sockets and sockscan.
# Used by x86 2003 SP1 and SP2 only.
tcpip_vtypes_2003_sp1_sp2 = {
    '_ADDRESS_OBJECT' : [0x68, {
        'Next' : [0x0, ['pointer', ['_ADDRESS_OBJECT']]],
        'LocalIpAddress' : [0x30, ['Ipv4Address']],
        'LocalPort' : [0x34, ['unsigned be short']],
        'Protocol'  : [0x36, ['unsigned short']],
        'Pid' : [0x14C, ['unsigned long']],
        'CreateTime' : [0x158, ['WinFileTime', {}]],
    }],
}

TCP_STATE_ENUM = {
    0: 'CLOSED', 1: 'LISTENING', 2: 'SYN_SENT',
    3: 'SYN_RCVD', 4: 'ESTABLISHED', 5: 'FIN_WAIT1',
    6: 'FIN_WAIT2', 7: 'CLOSE_WAIT', 8: 'CLOSING',
    9: 'LAST_ACK', 12: 'TIME_WAIT', 13: 'DELETE_TCB'
    }

# Structures used by netscan for x86 Vista and 2008 (all service packs).
tcpip_vtypes_vista = {
    '_IN_ADDR' : [None, {
        'addr4' : [0x0, ['Ipv4Address']],
        'addr6' : [0x0, ['Ipv6Address']],
    }],
    '_LOCAL_ADDRESS' : [None, {
        'pData' : [0xC, ['pointer', ['pointer', ['_IN_ADDR']]]],
    }],
    '_TCP_LISTENER': [0xa8, { # TcpL
        'Owner' : [0x18, ['pointer', ['_EPROCESS']]],
        'CreateTime' : [0x20, ['WinFileTime', {}]],
        'LocalAddr' : [0x34, ['pointer', ['_LOCAL_ADDRESS']]],
        'InetAF' : [0x38, ['pointer', ['_INETAF']]],
        "Endpoint": [0x50, ['Pointer', dict(
            target="_TCP_ENDPOINT"
        )]],
        'Port' : [0x3E, ['unsigned be short']],
    }],
    '_TCP_ENDPOINT': [0x1f0, { # TcpE
        'InetAF' : [0xC, ['pointer', ['_INETAF']]],
        'AddrInfo' : [0x10, ['pointer', ['_ADDRINFO']]],
        'ListEntry': [0x14, ['_LIST_ENTRY']],
        'State' : [0x28, ['Enumeration', dict(
            target='long',
            choices=TCP_STATE_ENUM)]],
        'LocalPort' : [0x2C, ['unsigned be short']],
        'RemotePort' : [0x2E, ['unsigned be short']],
        'Owner' : [0x160, ['pointer', ['_EPROCESS']]],
        'CreateTime' : [0, ['WinFileTime', {}]],
    }],
    '_TCP_SYN_ENDPOINT': [None, {
        'ListEntry': [8, ['_LIST_ENTRY']],
        'InetAF' : [0x18, ['pointer', ['_INETAF']]],
        'LocalPort' : [0x3c, ['unsigned be short']],
        'RemotePort' : [0x3e, ['unsigned be short']],
        'LocalAddr' : [0x1c, ['pointer', ['_LOCAL_ADDRESS']]],
        'RemoteAddress' : [0x28, ['pointer', ['_IN_ADDR']]],
        'Owner' : [0x20, ['pointer', ['_SYN_OWNER']]],
        'CreateTime' : [0, ['WinFileTime', {}]],
    }],
    '_SYN_OWNER': [None, {
        'Process': [0x18, ['pointer', ['_EPROCESS']]],
    }],
    '_TCP_TIMEWAIT_ENDPOINT': [None, {
        'ListEntry': [0x14, ['_LIST_ENTRY']],
        'InetAF' : [0xc, ['pointer', ['_INETAF']]],
        'LocalPort' : [0x1c, ['unsigned be short']],
        'RemotePort' : [0x1e, ['unsigned be short']],
        'LocalAddr' : [0x20, ['pointer', ['_LOCAL_ADDRESS']]],
        'RemoteAddress' : [0x24, ['pointer', ['_IN_ADDR']]],
        'CreateTime' : [0, ['WinFileTime', {}]],
    }],
    '_INETAF' : [None, {
        'AddressFamily' : [0xC, ['unsigned short']],
    }],
    '_ADDRINFO' : [None, {
        'Local' : [0x0, ['pointer', ['_LOCAL_ADDRESS']]],
        'Remote' : [0x8, ['pointer', ['_IN_ADDR']]],
    }],
    '_UDP_ENDPOINT': [0xa8, { # UdpA
        'Owner' : [0x18, ['pointer', ['_EPROCESS']]],
        'CreateTime' : [0x30, ['WinFileTime', {}]],
        'LocalAddr' : [0x38, ['pointer', ['_LOCAL_ADDRESS']]],
        'InetAF' : [0x14, ['pointer', ['_INETAF']]],
        'Port' : [0x48, ['unsigned be short']],
    }],
}

# Structures for netscan on x86 Windows 7 (all service packs).
tcpip_vtypes_7 = {
    '_TCP_ENDPOINT': [0x210, { # TcpE
        'InetAF' : [0xC, ['pointer', ['_INETAF']]],
        'AddrInfo' : [0x10, ['pointer', ['_ADDRINFO']]],
        'ListEntry': [0x14, ['_LIST_ENTRY']],
        'State' : [0x34, ['Enumeration', dict(
            target='long', choices=TCP_STATE_ENUM)]],
        'LocalPort' : [0x38, ['unsigned be short']],
        'RemotePort' : [0x3A, ['unsigned be short']],
        'Owner' : [0x174, ['pointer', ['_EPROCESS']]],
        'CreateTime' : [0, ['WinFileTime', {}]],
    }],
    '_TCP_SYN_ENDPOINT': [None, {
        'ListEntry': [8, ['_LIST_ENTRY']],
        'InetAF' : [0x24, ['pointer', ['_INETAF']]],
        'LocalPort' : [0x48, ['unsigned be short']],
        'RemotePort' : [0x4a, ['unsigned be short']],
        'LocalAddr' : [0x28, ['pointer', ['_LOCAL_ADDRESS']]],
        'RemoteAddress' : [0x34, ['pointer', ['_IN_ADDR']]],
        'Owner' : [0x2c, ['pointer', ['_SYN_OWNER']]],
        'CreateTime' : [0, ['WinFileTime', {}]],
    }],
    '_TCP_TIMEWAIT_ENDPOINT': [None, {
        'ListEntry': [0, ['_LIST_ENTRY']],
        'InetAF' : [0x18, ['pointer', ['_INETAF']]],
        'LocalPort' : [0x28, ['unsigned be short']],
        'RemotePort' : [0x2a, ['unsigned be short']],
        'LocalAddr' : [0x2c, ['pointer', ['_LOCAL_ADDRESS']]],
        'RemoteAddress' : [0x30, ['pointer', ['_IN_ADDR']]],
        'CreateTime' : [0, ['WinFileTime', {}]],
    }],
}

# Structures for netscan on x64 Vista SP0 and 2008 SP0
tcpip_vtypes_vista_64 = {
    '_IN_ADDR' : [None, {
        'addr4' : [0x0, ['Ipv4Address']],
        'addr6' : [0x0, ['Ipv6Address']],
    }],
    '_TCP_LISTENER': [0x120, { # TcpL
        'Owner' : [0x28, ['pointer', ['_EPROCESS']]],
        'CreateTime' : [0x20, ['WinFileTime', {}]],
        'LocalAddr' : [0x58, ['pointer', ['_LOCAL_ADDRESS']]],
        'InetAF' : [0x60, ['pointer', ['_INETAF']]],
        'Port' : [0x6a, ['unsigned be short']],
    }],
    '_INETAF' : [None, {
        'AddressFamily' : [0x14, ['unsigned short']],
    }],
    '_LOCAL_ADDRESS' : [None, {
        'pData' : [0x10, ['pointer', ['pointer', ['_IN_ADDR']]]],
    }],
    '_ADDRINFO' : [None, {
        'Local' : [0x0, ['pointer', ['_LOCAL_ADDRESS']]],
        'Remote' : [0x10, ['pointer', ['_IN_ADDR']]],
    }],
    '_TCP_ENDPOINT': [0x210, { # TcpE
        'InetAF' : [0x18, ['pointer', ['_INETAF']]],
        'AddrInfo' : [0x20, ['pointer', ['_ADDRINFO']]],
        'ListEntry': [0x28, ['_LIST_ENTRY']],
        'State' : [0x50, ['Enumeration', dict(
            target='long',
            choices=TCP_STATE_ENUM)]],
        'LocalPort' : [0x54, ['unsigned be short']],
        'RemotePort' : [0x56, ['unsigned be short']],
        'Owner' : [0x208, ['pointer', ['_EPROCESS']]],
        'CreateTime' : [0, ['WinFileTime', {}]],
    }],
    '_TCP_SYN_ENDPOINT': [None, {
        'ListEntry': [0x10, ['_LIST_ENTRY']],
        'InetAF' : [0x30, ['pointer', ['_INETAF']]],
        'LocalPort' : [0x64, ['unsigned be short']],
        'RemotePort' : [0x66, ['unsigned be short']],
        'LocalAddr' : [0x38, ['pointer', ['_LOCAL_ADDRESS']]],
        'RemoteAddress' : [0x50, ['pointer', ['_IN_ADDR']]],
        'Owner' : [0x40, ['pointer', ['_SYN_OWNER']]],
        'CreateTime' : [0, ['WinFileTime', {}]],
    }],
    '_SYN_OWNER': [None, {
        'Process': [0x28, ['pointer', ['_EPROCESS']]],
    }],
    '_TCP_TIMEWAIT_ENDPOINT': [None, {
        'ListEntry': [0, ['_LIST_ENTRY']],
        'InetAF' : [0x18, ['pointer', ['_INETAF']]],
        'LocalPort' : [0x30, ['unsigned be short']],
        'RemotePort' : [0x32, ['unsigned be short']],
        'LocalAddr' : [0x38, ['pointer', ['_LOCAL_ADDRESS']]],
        'RemoteAddress' : [0x40, ['pointer', ['_IN_ADDR']]],
        'CreateTime' : [0, ['WinFileTime', {}]],
    }],
    '_UDP_ENDPOINT': [0x82, { # UdpA
        'Owner' : [0x28, ['pointer', ['_EPROCESS']]],
        'CreateTime' : [0x58, ['WinFileTime', {}]],
        'LocalAddr' : [0x60, ['pointer', ['_LOCAL_ADDRESS']]],
        'InetAF' : [0x20, ['pointer', ['_INETAF']]],
        'Port' : [0x80, ['unsigned be short']],
    }],
    }


tcpip_vtypes_win7_64 = {
    '_TCP_ENDPOINT': [0x320, {
        'State' : [0x68, ['Enumeration', dict(target='long',
                                              choices=TCP_STATE_ENUM)]],
        'LocalPort' : [0x6c, ['unsigned be short']],
        'RemotePort' : [0x6e, ['unsigned be short']],
        'Owner' : [0x238, ['pointer', ['_EPROCESS']]],
    }],
    '_TCP_SYN_ENDPOINT': [None, {
        'InetAF' : [0x48, ['pointer', ['_INETAF']]],
        'LocalPort' : [0x7c, ['unsigned be short']],
        'RemotePort' : [0x7e, ['unsigned be short']],
        'LocalAddr' : [0x50, ['pointer', ['_LOCAL_ADDRESS']]],
        'RemoteAddress' : [0x68, ['pointer', ['_IN_ADDR']]],
        'Owner' : [0x58, ['pointer', ['_SYN_OWNER']]],
    }],
    '_TCP_TIMEWAIT_ENDPOINT': [None, {
        'InetAF' : [0x30, ['pointer', ['_INETAF']]],
        'LocalPort' : [0x48, ['unsigned be short']],
        'RemotePort' : [0x4a, ['unsigned be short']],
        'LocalAddr' : [0x50, ['pointer', ['_LOCAL_ADDRESS']]],
        'RemoteAddress' : [0x58, ['pointer', ['_IN_ADDR']]],
    }],
    }



overlays = {
    "_ADDRESS_OBJECT": [None, {
        "Protocol": [None, ["Enumeration", dict(
            choices=protos,
            target="unsigned int")]],
    }]
}


# This is not used just yet but soon!
win7_x86_dynamic_overlays = {
    "_TCP_LISTENER": dict(
        # The Owner process.
        Owner=[
            # Attempt 1
            ["Disassembler", dict(
                start="tcpip.sys!_TcpCreateListener@8",
                length=300,
                rules=[
                    "CALL *InetGetClientProcess",
                    "MOV [EBX+$out], EAX",
                ],
                target="Pointer",
                target_args=dict(
                    target="_EPROCESS"
                ),
            )],

            # Attempt 2
            ["Disassembler", dict(
                start="tcpip.sys!_TcpCovetNetBufferList@20",
                rules=[
                    "MOV EAX, [ESI+$out]",
                    "TEST EAX, EAX",
                    "PUSH EAX",
                    "CALL DWORD *PsGetProcessId",
                ],
                target="Pointer",
                target_args=dict(
                    target="_EPROCESS"
                ),
            )]
        ],
        # Socket creation time.
        CreateTime=[
            ["Disassembler", dict(
                start="tcpip.sys!_TcpCreateListener@8",
                length=300,
                rules=[
                    "LEA EAX, [EBX+$out]",
                    "PUSH EAX",
                    "CALL DWORD *KeQuerySystemTime",
                ],
                target="WinFileTime",
            )],
        ],
    ),
}



class _TCP_LISTENER(obj.Struct):
    """Class for objects found in TcpL pools"""

    def dual_stack_sockets(self, vm=None):
        """Handle Windows dual-stack sockets"""

        # If this pointer is valid, the socket is bound to
        # a specific IP address. Otherwise, the socket is
        # listening on all IP addresses of the address family.
        local_addr = self.LocalAddr.dereference(vm=vm)

        # Switch to the correct address space.
        af_inet = self.InetAF.dereference(vm=vm)

        # Note the remote address is always INADDR_ANY or
        # INADDR6_ANY for sockets. The moment a client
        # connects to the listener, a TCP_ENDPOINT is created
        # and that structure contains the remote address.
        if local_addr:
            inaddr = local_addr.pData.dereference().dereference()
            if af_inet.AddressFamily == AF_INET:
                yield "v4", inaddr.addr4, inaddr_any
            else:
                yield "v6", inaddr.addr6, inaddr6_any
        else:
            yield "v4", inaddr_any, inaddr_any
            if af_inet.AddressFamily.v() == AF_INET6:
                yield "v6", inaddr6_any, inaddr6_any


class _TCP_ENDPOINT(_TCP_LISTENER):
    """Class for objects found in TcpE pools"""

    def _ipv4_or_ipv6(self, in_addr, vm=None):
        af_inet = self.InetAF.dereference(vm=vm)
        if af_inet.AddressFamily == AF_INET:
            return in_addr.addr4
        else:
            return in_addr.addr6

    def LocalAddress(self, vm=None):
        inaddr = self.AddrInfo.dereference(vm=vm).Local.\
            pData.dereference().dereference()

        return self._ipv4_or_ipv6(inaddr, vm=vm)

    def RemoteAddress(self, vm=None):
        inaddr = self.AddrInfo.dereference(vm=vm).\
            Remote.dereference()

        return self._ipv4_or_ipv6(inaddr, vm=vm)


class _UDP_ENDPOINT(_TCP_LISTENER):
    """Class for objects found in UdpA pools"""


class TcpipPluginMixin(object):
    """A mixin for plugins that want to use tcpip.sys profiles."""

    @classmethod
    def args(cls, parser):
        super(TcpipPluginMixin, cls).args(parser)
        parser.add_argument("--tcpip_guid", default=None,
                            help="Force this profile to be used for tcpip.")

    def __init__(self, tcpip_guid=None, **kwargs):
        super(TcpipPluginMixin, self).__init__(**kwargs)
        # For the address resolver to load this GUID.
        if tcpip_guid:
            self.session.SetCache("tcpip_guid", tcpip_guid)

        self.tcpip_profile = self.session.address_resolver.LoadProfileForName(
            "tcpip")

        if not self.tcpip_profile:
            raise RuntimeError("Unable to load the profile for tcpip.sys")


class Tcpip(pe_vtypes.BasicPEProfile):
    """A profile for the TCPIP driver."""

    @classmethod
    def Initialize(cls, profile):
        super(Tcpip, cls).Initialize(profile)

        # Network Object Classess for Vista, 2008, and 7 x86 and x64
        if profile.get_constant("TCP_LISTENER_ACTIVATED"):
            profile.add_classes(dict(_TCP_LISTENER=_TCP_LISTENER,
                                     _TCP_ENDPOINT=_TCP_ENDPOINT,
                                     _UDP_ENDPOINT=_UDP_ENDPOINT))

        # Switch on the kernel version. FIXME: This should be done using the
        # generate_types module.
        version = profile.session.profile.metadata("version")

        if profile.metadata("arch") == "AMD64":
            # Vista SP1.
            if version == 6.0:
                profile.add_overlay(tcpip_vtypes_vista_64)
                profile.add_overlay({
                    '_TCP_ENDPOINT': [None, {
                        'Owner' : [0x210, ['pointer', ['_EPROCESS']]],
                        }],
                    })

            # Windows 7
            elif version >= 6.1:
                profile.add_overlay(tcpip_vtypes_vista_64)
                profile.add_overlay(tcpip_vtypes_win7_64)

            # Win2k3
            elif version == 5.2:
                profile.add_overlay(tcpip_vtypes_2003_x64)

        elif profile.metadata("arch") == "I386":
            profile.add_overlay(tcpip_vtypes)

            # Win2k3
            if version == 5.2:
                profile.add_overlay(tcpip_vtypes_2003_sp1_sp2)

            # Vista
            elif version == 6.0:
                profile.add_overlay(tcpip_vtypes_vista)

            # Windows 7
            elif version >= 6.1:
                profile.add_overlay(tcpip_vtypes_vista)
                profile.add_overlay(tcpip_vtypes_7)

        # Pool tags
        profile.add_constants(UDP_END_POINT_POOLTAG="UdpA",
                              TCP_LISTENER_POOLTAG="TcpL",
                              TCP_END_POINT_POOLTAG="TcpE")


        profile.add_overlay(overlays)

        return profile


class TcpipHook(kb.ParameterHook):
    name = "tcpip_profile"

    def calculate(self):
        index = self.session.LoadProfile("tcpip/index")
        image_base = self.session.address_resolver.get_address_by_name("tcpip")

        for guess, _ in index.LookupIndex(image_base):
            return guess



# Rekall Memory Forensics
# Copyright (c) 2012, Michael Cohen <scudette@gmail.com>
# Copyright (c) 2010, 2011, 2012 Michael Ligh <michael.ligh@mnin.org>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#


"""A Rekall Memory Forensics scanner which uses yara."""
from rekall import scan


class BaseYaraASScanner(scan.BaseScanner):
    """An address space scanner for Yara signatures."""
    overlap = 1024

    def __init__(self, rules=None, **kwargs):
        super(BaseYaraASScanner, self).__init__(**kwargs)
        self.rules = rules
        self.hits = []
        self.base_offset = None

    def _match_rules(self, buffer_as):
        """Compatibility for yara modules.

        Unfortunately there are two different implementations of the yara python
        bindings:

        # The original upstream source.
        http://plusvic.github.io/yara/

        # The version which is installed using pip install.
        https://github.com/mjdorma/yara-ctypes

        These do not work the same and so we need to support both.

        Yields:
          a tuple of (offset, rule_name, name, value)
        """
        matches = self.rules.match(data=buffer_as.data)
        # yara-cpython bindings from pip.
        if type(matches) is dict:
            for _, matches in matches.items():
                for match in matches:
                    for string in match["strings"]:
                        hit_offset = string["offset"] + buffer_as.base_offset

                        yield (match["rule"], hit_offset,
                               string["identifier"], string["data"])

        else:
            # native bindings from http://plusvic.github.io/yara/
            for match in matches:
                for buffer_offset, name, value in match.strings:
                    hit_offset = buffer_offset + buffer_as.base_offset
                    yield (match.rule, hit_offset, name, value)

    def check_addr(self, scan_offset, buffer_as=None):
        # The buffer was changed - we scan the entire buffer and record the
        # hits - then we can feed it to the Rekall scan framework.
        if self.base_offset != buffer_as.base_offset:
            self.base_offset = buffer_as.base_offset
            self.hits = []

            for rule, offset, name, value in self._match_rules(buffer_as):
                self.hits.append((rule, offset, name, value))

        if self.hits and scan_offset == self.hits[0][1]:
            return self.hits.pop(0)

    def skip(self, buffer_as, offset):
        # Skip the entire buffer.
        if not self.hits:
            return len(buffer_as.data)

        next_hit = self.hits[0][1]
        return next_hit - offset



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# pylint: disable=protected-access

from rekall import obj
from rekall import scan
from rekall import plugin
from rekall.plugins.windows import common


class KDBGScanner(scan.BaseScanner):
    """Scans for _KDDEBUGGER_DATA64 structures.

    Note that this does not rely on signatures, as validity of hits is
    calculated through list reflection.
    """
    checks = [("StringCheck", dict(needle="KDBG"))]

    def scan(self, offset=0, maxlen=None):
        maxlen = maxlen or self.profile.get_constant("MaxPointer")

        # How far into the struct the OwnerTag is.
        owner_tag_offset = self.profile.get_obj_offset(
            "_DBGKD_DEBUG_DATA_HEADER64", "OwnerTag")

        # Depending on the memory model this behaves slightly differently.
        architecture = self.profile.metadata("arch", "I386")

        # This basically iterates over all hits on the string "KDBG".
        for offset in super(KDBGScanner, self).scan(offset, maxlen):
            # For each hit we overlay a _DBGKD_DEBUG_DATA_HEADER64 on it and
            # reflect through the "List" member.
            result = self.profile.Object("_KDDEBUGGER_DATA64",
                                         offset=offset - owner_tag_offset,
                                         vm=self.address_space)

            # We verify this hit by reflecting through its header list.
            list_entry = result.Header.List

            # On 32 bit systems the Header.List member seems to actually be a
            # LIST_ENTRY32 instead of a LIST_ENTRY64, but it is still padded to
            # take the same space:
            if architecture == "I386":
                list_entry = list_entry.cast("LIST_ENTRY32")

            if list_entry.reflect():
                yield result

            elif (list_entry.Flink == list_entry.Blink and
                  not list_entry.Flink.dereference()):
                self.session.logging.debug(
                    "KDBG list_head is not mapped, assuming its valid.")

                yield result


class KDBGScan(plugin.KernelASMixin, common.AbstractWindowsCommandPlugin):
    """Scan for possible _KDDEBUGGER_DATA64 structures.

    The scanner is detailed here:
    http://moyix.blogspot.com/2008/04/finding-kernel-global-variables-in.html

    The relevant structures are detailed here:
    http://doxygen.reactos.org/d3/ddf/include_2psdk_2wdbgexts_8h_source.html

    We can see that _KDDEBUGGER_DATA64.Header is:

    typedef struct _DBGKD_DEBUG_DATA_HEADER64 {
        LIST_ENTRY64    List;
        ULONG           OwnerTag;
        ULONG           Size;
    }

    We essentially search for an owner tag of "KDBG", then overlay the
    _KDDEBUGGER_DATA64 struct on it. We test for validity by reflecting
    through the Header.List member.
    """

    __name = "kdbgscan"

    @classmethod
    def args(cls, parser):
        super(KDBGScan, cls).args(parser)
        parser.add_argument("--full_scan", default=False, type="Boolean",
                            help="Scan the full address space.")


    def __init__(self, full_scan=False, **kwargs):
        super(KDBGScan, self).__init__(**kwargs)
        self.full_scan = full_scan

    def hits(self):
        if self.full_scan:
            start, end = 0, 2**64
        else:
            # The kernel image is always loaded in the same range called the
            # "Initial Loader Mappings". Narrowing the possible range makes
            # scanning much faster. (See
            # http://www.codemachine.com/article_x64kvas.html)
            if self.session.profile.metadata("arch") == "AMD64":
                start, end = 0xFFFFF80000000000, 0xFFFFF87FFFFFFFFF
            else:
                start, end = 0x80000000, 0xFFFFFFFF

        scanner = KDBGScanner(
            session=self.session, profile=self.profile,
            address_space=self.kernel_address_space)

        # Yield actual objects here
        for kdbg in scanner.scan(
                obj.Pointer.integer_to_address(start),
                end - start):
            yield kdbg

    def render(self, renderer):
        """Renders the KPCR values as text"""

        for kdbg in self.hits():
            renderer.section()
            renderer.format(
                "Instantiating KDBG using: {0} {1} ({2})\n",
                kdbg.obj_vm.name, kdbg.obj_profile.__class__.__name__,
                kdbg.obj_profile.metadata('arch', "Unknown"),
                )

            renderer.format("{0:<30}: {1:#x}\n", "Offset (V)", kdbg.obj_offset)
            renderer.format("{0:<30}: {1:#x}\n", "Offset (P)", kdbg.obj_vm.vtop(
                kdbg.obj_offset))

            # These fields can be gathered without dereferencing
            # any pointers, thus they're available always
            renderer.format("{0:<30}: {1}\n", "KDBG owner tag check",
                            kdbg.is_valid())

            verinfo = kdbg.dbgkd_version64()
            if verinfo:
                renderer.format(
                    "{0:<30}: {1:#x} (Major: {2}, Minor: {3})\n",
                    "Version64", verinfo.obj_offset, verinfo.MajorVersion,
                    verinfo.MinorVersion)

            renderer.format("{0:<30}: {1}\n", "Service Pack (CmNtCSDVersion)",
                            kdbg.ServicePack)

            renderer.format("{0:<30}: {1}\n", "Build string (NtBuildLab)",
                            kdbg.NtBuildLab.dereference())

            # Count the total number of tasks from PsActiveProcessHead.
            try:

                pslist = kdbg.PsActiveProcessHead.list_of_type(
                    "_EPROCESS", "ActiveProcessLinks")
                num_tasks = len([x for x in pslist if x.pid > 0])
            except AttributeError:
                num_tasks = 0

            try:
                modules = self.session.plugins.modules(session=self.session)
                num_modules = len(list(modules.lsmod()))
            except AttributeError:
                num_modules = 0

            renderer.format("{0:<30}: {1:#x} ({2} processes)\n",
                            "PsActiveProcessHead",
                            kdbg.PsActiveProcessHead, num_tasks)

            renderer.format("{0:<30}: {1:#x} ({2} modules)\n",
                            "PsLoadedModuleList",
                            kdbg.PsLoadedModuleList, num_modules)

            renderer.format("{0:<30}: {1:#x} (Matches MZ: {2})\n",
                            "KernelBase", kdbg.KernBase,
                            kdbg.obj_vm.read(kdbg.KernBase, 2) == "MZ")

            # Parse the PE header of the kernel.
            pe_profile = self.session.LoadProfile("pe")

            dos_header = pe_profile._IMAGE_DOS_HEADER(
                offset=kdbg.KernBase, vm=kdbg.obj_vm)
            nt_header = dos_header.NTHeader
            if nt_header:
                renderer.format(
                    "{0:<30}: {1}\n", "Major (OptionalHeader)",
                    nt_header.OptionalHeader.MajorOperatingSystemVersion)
                renderer.format(
                    "{0:<30}: {1}\n", "Minor (OptionalHeader)",
                    nt_header.OptionalHeader.MinorOperatingSystemVersion)

            # The CPU block.
            for kpcr in kdbg.kpcrs():
                renderer.format(
                    "{0:<30}: {1:#x} (CPU {2})\n",
                    "KPCR", kpcr.obj_offset, kpcr.ProcessorBlock.Number)



# Rekall Memory Forensics
#
# Based on the source code from
# Volatools Basic
# Copyright (C) 2007 Komoku, Inc.
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@gmail.com>
# Brendan Dolan-Gavitt <bdolangavitt@wesleyan.edu>
# Mike Auty <mike.auty@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# The source code in this file was inspired by the excellent work of
# Brendan Dolan-Gavitt. Background information can be found in
# the following reference:
# "The VAD Tree: A Process-Eye View of Physical Memory," Brendan Dolan-Gavitt

# pylint: disable=protected-access

import re

from rekall import scan
from rekall import utils
from rekall.plugins import core
from rekall.plugins.addrspaces import intel
from rekall.plugins.common import pfn
from rekall.plugins.windows import common
from rekall.plugins.windows import pagefile


class VADInfo(common.WinProcessFilter):
    """Dump the VAD info"""

    __name = "vadinfo"

    def render(self, renderer):
        for task in self.filter_processes():
            renderer.section()
            renderer.format("Pid: {0:6}\n", task.UniqueProcessId)

            count = 0
            for count, vad in enumerate(task.RealVadRoot.traverse()):
                try:
                    self.write_vad_short(renderer, vad)
                except AttributeError:
                    pass

                try:
                    self.write_vad_control(renderer, vad)
                except AttributeError:
                    pass

                try:
                    self.write_vad_ext(renderer, vad)
                except AttributeError:
                    pass

                renderer.format("\n")

            self.session.report_progress("Pid %s: %s Vads" % (
                task.UniqueProcessId, count))

    def write_vad_short(self, renderer, vad):
        """Renders a text version of a Short Vad"""
        renderer.table_header([("VAD node @", "offset", ""),
                               ("address", "address", "[addrpad]"),
                               ("Start", "Start", "5"),
                               ("startaddr", "startaddr", "[addrpad]"),
                               ("End", "End", "3"),
                               ("endaddr", "endaddr", "[addrpad]"),
                               ("Tag", "Tag", "3"),
                               ("tagval", "tagval", ""),
                              ], suppress_headers=True)
        renderer.table_row("VAD node @",
                           vad.obj_offset,
                           "Start",
                           vad.Start,
                           "End",
                           vad.End,
                           "Tag",
                           vad.Tag)

        renderer.format("Flags: {0}\n", vad.u.VadFlags)

        # although the numeric value of Protection is printed above with
        # VadFlags, let's show the user a human-readable translation of the
        # protection
        renderer.format("Protection: {0}\n", vad.u.VadFlags.ProtectionEnum)

        # translate the vad type if its available (> XP)
        if vad.u.VadFlags.m("VadType"):
            renderer.format("Vad Type: {0}\n", vad.u.VadFlags.VadTypeEnum)

    def write_vad_control(self, renderer, vad):
        """Renders a text version of a (non-short) Vad's control information"""
        # even if the ControlArea is not NULL, it is only meaningful
        # for shared (non private) memory sections.
        if vad.u.VadFlags.PrivateMemory == 1:
            return

        control_area = vad.ControlArea
        if not control_area:
            return

        renderer.format("ControlArea @{0:08x} Segment {1:08x}\n",
                        control_area.dereference().obj_offset,
                        control_area.Segment)

        renderer.format("Dereference list: Flink {0:08x}, Blink {1:08x}\n",
                        control_area.DereferenceList.Flink,
                        control_area.DereferenceList.Blink)

        renderer.format(
            "NumberOfSectionReferences: {0:10} NumberOfPfnReferences:  "
            "{1:10}\n", control_area.NumberOfSectionReferences,
            control_area.NumberOfPfnReferences)

        renderer.format(
            "NumberOfMappedViews:       {0:10} NumberOfUserReferences: "
            "{1:10}\n", control_area.NumberOfMappedViews,
            control_area.NumberOfUserReferences)

        renderer.format(
            "WaitingForDeletion Event:  {0:08x}\n",
            control_area.WaitingForDeletion)

        renderer.format(
            "Control Flags: {0}\n", control_area.u.Flags)

        file_object = control_area.FilePointer
        if file_object and file_object != 0:
            renderer.format(
                "FileObject @{0:08x} FileBuffer @ {1:08x}          , "
                "Name: {2}\n", file_object.obj_offset,
                file_object.FileName.Buffer, file_object.FileName)

    def write_vad_ext(self, renderer, vad):
        """Renders a text version of a Long Vad"""
        if vad.obj_type != "_MMVAD_SHORT":
            renderer.format(
                "First prototype PTE: {0:08x} Last contiguous PTE: "
                "{1:08x}\n", vad.FirstPrototypePte[0],
                vad.LastContiguousPte)

            renderer.format("Flags2: {0}\n", vad.u2.VadFlags2)



class VADTree(VADInfo):
    """Walk the VAD tree and display in tree format"""

    __name = "vadtree"

    def render(self, renderer):
        for task in self.filter_processes():
            renderer.section()
            renderer.format(u"Pid: {0:6}\n", task.UniqueProcessId)

            renderer.table_header([
                dict(name="", type="TreeNode", width=1),
                ("Start", "start", "[addrpad]"),
                ("-", "-", ""),
                ("End", "End", "[addrpad]")
                ], suppress_headers=True)

            for vad in task.RealVadRoot.traverse():
                level = vad.obj_context.get('depth', 0)
                renderer.table_row("", vad.Start, "->", vad.End, depth=level)


class VADWalk(VADInfo):
    """Walk the VAD tree"""

    __name = "vadwalk"

    def render(self, renderer):
        for task in self.filter_processes():
            renderer.section()
            renderer.format(u"Pid: {0:6}\n", task.UniqueProcessId)
            renderer.table_header([("Address", "address", "[addrpad]"),
                                   ("Parent", "parent", "[addrpad]"),
                                   ("Left", "left", "[addrpad]"),
                                   ("Right", "right", "[addrpad]"),
                                   ("Start", "start", "[addrpad]"),
                                   ("End", "end", "[addrpad]"),
                                   ("Tag", "tag", "4"),
                                  ])
            for vad in task.RealVadRoot.traverse():
                # Ignore Vads with bad tags (which we explicitly include as
                # None)
                if vad:
                    renderer.table_row(
                        vad.obj_offset,
                        vad.obj_parent.obj_offset,
                        vad.LeftChild.v(),
                        vad.RightChild.v(),
                        vad.Start,
                        vad.End,
                        vad.Tag)

class VAD(common.WinProcessFilter):
    """Concise dump of the VAD.

    Similar to windbg's !vad.
    """

    __name = "vad"

    @classmethod
    def args(cls, parser):
        super(VAD, cls).args(parser)
        parser.add_argument(
            "--regex", default=None,
            help="A regular expression to filter VAD filenames.")

        parser.add_argument(
            "--offset", default=None, type="IntParser",
            help="Only print the vad corresponding to this offset.")

    def __init__(self, *args, **kwargs):
        self.regex = kwargs.pop("regex", None)
        self.offset = kwargs.pop("offset", None)
        self._cache = {}

        # Pass positional args to the WinProcessFilter constructor.
        super(VAD, self).__init__(*args, **kwargs)

    def find_file(self, addr):
        """Finds the file mapped at this address."""
        for task in self.filter_processes():
            yield self.find_file_in_task(addr, task)

    def find_file_in_task(self, addr, task):
        resolver = self.GetVadsForProcess(task)
        try:
            if resolver:
                return resolver.find_le(addr)
        except ValueError:
            return None

    def _make_cache(self, task):
        result = utils.SortedCollection(key=lambda x: x[0])
        self.session.report_progress(
            " Enumerating VADs in %s (%s)", task.name, task.pid)

        for vad in task.RealVadRoot.traverse():
            result.insert(
                (vad.Start, vad.End, self._get_filename(vad), vad))

        return result

    def GetVadsForProcess(self, task):
        try:
            resolver = self._cache[task]
        except KeyError:
            # Break the recursion by placing a None for the resolver. The
            # self._make_cache() call will run the vad plugin which might
            # resolve a VAD PTE, calling this code. This means that we can only
            # use non-VAD PTEs to read the VAD itself.
            self._cache[task] = None
            resolver = self._cache[task] = self._make_cache(task)

        return resolver

    def _get_filename(self, vad):
        filename = ""
        try:
            file_obj = vad.ControlArea.FilePointer
            if file_obj:
                filename = file_obj.FileName or "Pagefile-backed section"
        except AttributeError:
            pass

        return unicode(filename)

    def render_vadroot(self, renderer, vad_root, task):
        renderer.table_header([('VAD', 'offset', '[addrpad]'),
                               ('lev', 'level', '>3'),
                               ('Start Addr', 'start_pfn', '[addrpad]'),
                               ('End Addr', 'end_pfn', '[addrpad]'),
                               dict(name='com', width=6, align="r"),
                               ('', 'type', '7'),
                               ('', 'executable', '6'),
                               ('Protect', 'protection', '!20'),
                               ('Filename', 'filename', '')])

        task_as = task.get_process_address_space()

        result = []
        for vad in vad_root.traverse():
            # Apply filters if needed.
            if self.regex and not re.search(
                    self.regex, self._get_filename(vad)):
                continue

            if (self.offset is not None and
                    not vad.Start <= self.offset <= vad.End):
                continue

            exe = ""
            if "EXECUTE" in str(vad.u.VadFlags.ProtectionEnum):
                exe = "Exe"

            result.append((
                vad, vad.obj_context.get('depth', 0),

                # The vad region itself exists in the process address space.
                self.profile.Pointer(value=vad.Start, vm=task_as),
                self.profile.Pointer(value=vad.End, vm=task_as),
                vad.CommitCharge if vad.CommitCharge < 0x7fffffff else -1,
                "Private" if vad.u.VadFlags.PrivateMemory > 0 else "Mapped",
                exe,
                vad.u.VadFlags.ProtectionEnum,
                self._get_filename(vad)))

        # Sort by start range.
        result.sort(key=lambda x: x[2].v())
        for row in result:
            renderer.table_row(*row)

    def render(self, renderer):
        for task in self.filter_processes():
            renderer.section()
            renderer.format("Pid: {0} {1}\n", task.UniqueProcessId,
                            task.ImageFileName)
            renderer.RenderProgress("Pid: %s" % task.UniqueProcessId)
            self.render_vadroot(renderer, task.RealVadRoot, task)


class VADMap(pfn.VADMapMixin, common.WinProcessFilter):
    """Inspect each page in the VAD and report its status.

    This allows us to see the address translation status of each page in the
    VAD.
    """

    def _CreateMetadata(self, collection):
        metadata = {}
        for descriptor_cls, _, _ in collection.descriptors:
            # Any of the following kinds of PTE are considered prototype.
            if descriptor_cls in (pagefile.WindowsProtoTypePTEDescriptor,
                                  pagefile.VadPteDescriptor):
                metadata["ProtoType"] = True
                break

        for descriptor_cls, args, kwargs in reversed(collection.descriptors):
            if issubclass(descriptor_cls,
                          pagefile.WindowsSubsectionPTEDescriptor):
                metadata.update(descriptor_cls(
                    session=self.session, *args, **kwargs).metadata())

                return metadata

            elif issubclass(descriptor_cls, pagefile.WindowsPTEDescriptor):
                type = kwargs.get("pte_type")
                if type == "Proto":
                    metadata["ProtoType"] = True

                pte_value = kwargs.get("pte_value")
                if pte_value & 1:
                    metadata["type"] = "Valid"
                else:
                    metadata["type"] = "Transition"

                return metadata

            elif issubclass(descriptor_cls, pagefile.WindowsPagefileDescriptor):
                metadata["number"] = kwargs.get("pagefile_number", 0)
                metadata["offset"] = kwargs.get("address", 0)
                metadata["type"] = "Pagefile"

                return metadata

            elif issubclass(descriptor_cls, pagefile.DemandZeroDescriptor):
                metadata["type"] = "Demand Zero"
                return metadata

            elif issubclass(descriptor_cls, intel.PhysicalAddressDescriptor):
                metadata["offset"] = kwargs["address"]

        return metadata

    def _GenerateVADRuns(self, vad):
        address_space = self.session.GetParameter("default_address_space")

        offset = vad.Start
        end = vad.End

        while offset < end:
            if self.start <= offset <= self.end:
                yield offset, self._CreateMetadata(
                    address_space.describe_vtop(offset))

                self.session.report_progress("Inspecting 0x%08X", offset)

            offset += 0x1000

    def GeneratePageMetatadata(self, task):
        resolver = self.session.address_resolver
        for _, _, _, vad in resolver.GetVADs():
            # Skip guard regions for Wow64.
            if vad.CommitCharge >= 0x7fffffff:
                continue

            for vaddr, metadata in self._GenerateVADRuns(vad):
                yield vaddr, metadata


class VADDump(core.DirectoryDumperMixin, VAD):
    """Dumps out the vad sections to a file"""

    __name = "vaddump"

    @classmethod
    def args(cls, parser):
        super(VADDump, cls).args(parser)
        parser.add_argument("--max_size", type="IntParser",
                            default=100*1024*1024,
                            help="Maximum file size to dump.")

    def __init__(self, *args, **kwargs):
        self.max_size = kwargs.pop("max_size", 100*1024*1024)
        super(VADDump, self).__init__(*args, **kwargs)

    def render(self, renderer):
        for task in self.filter_processes():
            renderer.section("{0:6} ({1:2})".format(
                task.name, task.UniqueProcessId))

            renderer.table_header([
                ("Start", "start", "[addrpad]"),
                ("End", "end", "[addrpad]"),
                ("Length", "length", "[addr]"),
                ("Filename", "filename", "60s"),
                ("Comment", "comment", "")])

            # Get the task and all process specific information
            task_space = task.get_process_address_space()

            name = task.ImageFileName
            offset = task_space.vtop(task.obj_offset)
            if offset is None:
                renderer.format(
                    "Process does not have a valid address space.\n")
                continue

            with self.session.plugins.cc() as cc:
                cc.SwitchProcessContext(task)

                for vad in task.RealVadRoot.traverse():
                    # Find the start and end range
                    start = vad.Start
                    end = vad.End

                    filename = "{0}.{1:x}.{2:08x}-{3:08x}.dmp".format(
                        name, offset, start, end)

                    with renderer.open(directory=self.dump_dir,
                                       filename=filename,
                                       mode='wb') as fd:

                        if end - start > self.max_size:
                            renderer.table_row(start, end, end-start,
                                               "Skipped - Region too large")
                            continue

                        self.session.report_progress("Dumping %s" % filename)
                        self.CopyToFile(task_space, start, end + 1, fd)
                        renderer.table_row(
                            start, end, end-start, filename,
                            self._get_filename(vad))


class VadScanner(scan.BaseScanner):
    """A scanner over all memory regions of a process."""

    def __init__(self, task=None, process_profile=None, **kwargs):
        """Scan the process address space through the Vads.

        Args:
          task: The _EPROCESS object for this task.

          process_profile: The specialized profile for this process. In practice
            this is always different from task.obj_profile (which belongs to the
            kernel). If not provided we default to the kernel profile.
        """
        self.task = task
        super(VadScanner, self).__init__(
            profile=process_profile or task.obj_profile,
            address_space=task.get_process_address_space(),
            **kwargs)

    def scan(self, offset=0, maxlen=None):
        maxlen = maxlen or self.profile.get_constant("MaxPointer")

        for vad in self.task.RealVadRoot.traverse():
            # Only scan the VAD region.
            for match in super(VadScanner, self).scan(vad.Start, vad.Length):
                yield match



from rekall import testlib

from rekall.ui import text


class CellTest(testlib.RekallBaseUnitTestCase):

    def testNesting(self):
        c1 = text.Cell(value="Hello, world!", width=6, align="r")
        self.assertEqual(c1.lines, ["Hello,", "world!"])
        c1.rewrap(width=7)
        self.assertEqual(c1.lines, [" Hello,", " world!"])
        c1.rewrap(align="l")
        self.assertEqual(c1.lines, ["Hello, ", "world! "])

        c2 = text.Cell(value="I am a line of text.", width=13, align="l")
        self.assertEqual(c2.lines, ["I am a line  ", "of text.     "])
        c2.rewrap(width=8)
        self.assertEqual(c2.lines, ["I am a  ", "line of ", "text.   "])
        self.assertEqual(c2.width, 8)
        self.assertEqual(c2.height, 3)

        c3 = text.JoinedCell(c1, c2, tablesep="|")
        self.assertEqual(c3.lines, ["Hello, |I am a  ",
                                    "world! |line of ",
                                    "       |text.   "])
        self.assertEqual(c3.height, 3)
        self.assertEqual(c3.width, 7 + 1 + 8)

        c4 = text.JoinedCell(c3, c1)
        self.assertEqual(len(c4.cells), 3)
        self.assertEqual(c4.width, 7 + 1 + 8 + 1 + 7)

    def testColors(self):
        # Invisible characters shouldn't affect reported width.
        c1 = text.Cell(value="Hello, world!", highlights=[(0, 6, "RED", None)])
        self.assertEqual(c1.width, 13)

    def testPreserveNewLines(self):
        c1 = text.Cell(value="Hello,\n world!")
        self.assertEqual(c1.lines[0], "Hello, ")



import cPickle
import cStringIO
import os
import time

from rekall import config
from rekall import io_manager
from rekall import obj
from rekall import utils
from rekall.ui import json_renderer


config.DeclareOption(
    "--cache", default="file", type="String",
    choices=["file", "memory", "timed"],
    help="Type of cache to use. ")


class PicklingDirectoryIOManager(io_manager.DirectoryIOManager):
    def Encoder(self, data, **_):
        encoder = json_renderer.JsonEncoder(
            session=self.session, renderer="JsonRenderer")
        data = encoder.Encode(data)

        try:
            return cPickle.dumps(data, -1)
        except TypeError:
            raise io_manager.EncodeError("Unable to pickle data")

    def Decoder(self, raw):
        """Safe Unpickling.

        Unpickle only safe primitives like tuples, dicts and
        strings. Specifically does not allow arbitrary instances to be
        recovered.
        """
        unpickler = cPickle.Unpickler(cStringIO.StringIO(raw))
        unpickler.find_global = None

        json_renderer_obj = json_renderer.JsonRenderer(
            session=self.session)
        decoder = json_renderer.JsonDecoder(
            self.session, json_renderer_obj)

        try:
            decoded = unpickler.load()
        except Exception:
            raise io_manager.DecodeError("Unable to unpickle cached object")

        return decoder.Decode(decoded)


class Cache(object):
    def __init__(self):
        self.data = {}

    def Get(self, item, default=None):
        return self.data.get(item, default)

    def Set(self, item, value, volatile=True):
        _ = volatile
        if value is None:
            self.data.pop(item, None)
        else:
            self.data[item] = value

    def Clear(self):
        self.data.clear()

    def Flush(self):
        """Called to sync the cache to external storage if required."""

    def __str__(self):
        """Print the contents somewhat concisely."""
        result = []
        for k, v in self.data.iteritems():
            if isinstance(v, obj.BaseObject):
                v = repr(v)

            value = "\n  ".join(str(v).splitlines())
            if len(value) > 1000:
                value = "%s ..." % value[:1000]

            result.append("  %s = %s" % (k, value))

        return "{\n" + "\n".join(sorted(result)) + "\n}"


class TimedCache(Cache):
    """A limited time Cache.

    This is useful for live analysis to ensure that information is not stale.
    """
    expire_time = 5

    def Get(self, item, default=None):
        now = time.time()
        data, timestamp = self.data.get(item, (default, now))
        if timestamp + self.expire_time < now:
            del self.data[item]
            return default

        return data

    def Set(self, item, value, volatile=True):
        """Sets the item to the value.

        The value will be cached for the expiry time if it is volatile (by
        default). Non-volatile data will never expire.

        Even on a live system, we cache information which can not change for the
        life of the system (e.g. the profile or dtb values). These are marked
        non-volatile and will not be expired.
        """
        if value is None:
            self.data.pop(item, None)
        else:
            if volatile:
                now = time.time()
            else:
                now = 2**63

            self.data[item] = (value, now)

    def __str__(self):
        """Print the contents somewhat concisely."""
        result = []
        now = time.time()
        for k, (v, timestamp) in self.data.items():
            if timestamp + self.expire_time < now:
                self.data.pop(k)
                continue

            if isinstance(v, obj.BaseObject):
                v = repr(v)

            value = "\n  ".join(str(v).splitlines())
            if len(value) > 1000:
                value = "%s ..." % value[:1000]

            result.append("  %s = %s" % (k, value))

        return "{\n" + "\n".join(sorted(result)) + "\n}"


class FileCache(Cache):
    """A cache which syncs to a persistent on disk representation.
    """
    def __init__(self, session):
        super(FileCache, self).__init__()
        self._io_manager = None
        self.session = session
        self.fingerprint = None
        self.name = None

        # Record all the dirty cached keys.
        self.dirty = set()

    @property
    def io_manager(self):
        cache_dir = self.session.GetParameter("cache_dir", cached=False)
        if self._io_manager is None and cache_dir:
            # Cache dir may be specified relative to the home directory.
            if config.GetHomeDir():
                cache_dir = os.path.join(config.GetHomeDir(), cache_dir)

            if os.access(cache_dir, os.F_OK | os.R_OK | os.W_OK | os.X_OK):
                self._io_manager = PicklingDirectoryIOManager(
                    "%s/sessions" % cache_dir, session=self.session,
                    mode="w")

        return self._io_manager

    def SetName(self, name):
        self.name = name

    def SetFingerprint(self, fingerprint):
        name = fingerprint["hash"]
        if self.name != name and self.io_manager:
            indexes = self.io_manager.GetData("sessions/index") or {}
            indexes[name] = fingerprint["tests"]

            self.name = name
            self.io_manager.StoreData("sessions/index", indexes)

    def Get(self, item, default=None):
        if (self.io_manager and             # We are backing to a file.
                item not in self.data and   # Item not already cached in memory.
                item not in self.dirty):    # Item was not previously changed.
            try:
                data = self.io_manager.GetData(
                    "sessions/%s/%s" % (self.name, item),
                    default=self)
                if data is not self:
                    self.data[item] = data
            except Exception:
                self.session.logging.error(
                    "Unable to decode cached object %s", item)

        return super(FileCache, self).Get(item, default=default)

    def Set(self, item, value, volatile=True):
        super(FileCache, self).Set(item, value, volatile=volatile)
        self.dirty.add(item)

    def Flush(self):
        """Write out all dirty items at once."""
        if self.fingerprint is None and self.session.HasParameter("profile"):
            self.SetFingerprint(self.session.GetParameter("image_fingerprint"))

        if self.name and self.io_manager:
            # Save to disk the dirty items.
            for key, item in self.data.iteritems():
                if key in self.dirty or getattr(item, "dirty", False):
                    now = time.time()
                    self.io_manager.StoreData(
                        "sessions/%s/%s" % (self.name, key), item)
                    self.session.logging.debug("Flushed %s in %s" % (
                        key, (time.time() - now)))

            self.io_manager.FlushInventory()

    def DetectImage(self, address_space):
        if not self.io_manager:
            return

        session_index = self.io_manager.GetData("sessions/index")
        for name, tests in session_index.iteritems():
            item = SessionIndex(name, tests)
            if item.Test(address_space):
                self.SetName(item.name)

                # Force current data to be flushed to disk so we do not lose it.
                self.Flush()
                return item.name


class SessionIndex(object):
    def __init__(self, name, tests):
        self.name = name
        self.test = tests

    def Test(self, address_space):
        for offset, expected in self.test:
            expected = utils.SmartStr(expected)
            if (offset and expected !=
                    address_space.read(offset, len(expected))):
                return False

        return True



def Factory(session, cache_type):
    """Instantiate the most appropriate cache for this session."""
    if cache_type == "memory":
        return Cache()

    if cache_type == "timed":
        return TimedCache()

    if cache_type == "file":
        return FileCache(session)

    return Cache()



# Rekall Memory Forensics
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""The module implements the base class for address resolution."""

__author__ = "Michael Cohen <scudette@gmail.com>"

import re

from rekall import config
from rekall import obj


config.DeclareOption(
    "--name_resolution_strategies", default=["Module", "Symbol", "Export"],
    group="Interface", type="ChoiceArray",
    choices=["Module", "Symbol", "Export"])


class AddressResolverMixin(object):
    """The basic building block for constructing an address resolver plugin."""

    # The name of the plugin.
    name = "address_resolver"

    # The format of a symbol name. Used by get_address_by_name().
    ADDRESS_NAME_REGEX = re.compile(
        r"(?P<deref>[*])?"              # Pointer dereference.

        r"((?P<address>0x[0-9A-Fa-f]+)|" # Alternative - Either an address, or,

        r"(?P<module>[A-Za-z_0-9\.\\]+)" # Module name - can include extension
                                         # (.exe, .sys)

        r"!?"                           # ! separates module name from symbol
                                        # name.

        r"(?P<symbol>[^ +-]+)?"         # Symbol name.
        r")"                            # End alternative.

        r"(?P<op> *[+-] *)?"            # Possible arithmetic operator.
        r"(?P<offset>[0-9a-fA-Fx]+)?")  # Possible hex offset.

    def __init__(self, **kwargs):
        super(AddressResolverMixin, self).__init__(**kwargs)
        self.profiles = {}

    def Reset(self):
        self.profiles.clear()

    def NormalizeModuleName(self, module):
        try:
            module_name = module.name
        except AttributeError:
            module_name = module

        module_name = unicode(module_name)
        module_name = re.split(r"[/\\]", module_name)[-1]

        return module_name.lower()

    def _EnsureInitialized(self):
        """Initialize this address resolver."""

    def _ParseAddress(self, name):
        m = self.ADDRESS_NAME_REGEX.match(name)
        if m:
            capture = m.groupdict()
            if not capture.get("address"):
                module = capture.get("module")
                if not module:
                    raise TypeError("Module name not specified.")

                capture["module"] = self.NormalizeModuleName(module)

            if capture["op"] and not (capture["symbol"] or
                                      capture["address"] or
                                      capture["module"]):
                raise TypeError("Operator %s must have an operand." %
                                capture["op"])

            if capture["op"] and not (capture["symbol"] or capture["address"]):
                raise TypeError(
                    "Operator %s must operate on a symbol or address." %
                    capture["op"])

            return capture

        raise TypeError("Unable to parse %r as a symbol name" % name)

    def FindContainingModule(self, address):
        """Finds the name of the module containing the specified address.

        Returns:
          A tuple of start address, end address, module name
        """
        _ = address
        # If we dont know anything about the address just return Nones.
        return None, None, None

    def get_constant_object(self, name, target, **kwargs):
        """Instantiate the named constant with these args."""
        self._EnsureInitialized()

        profile = self.LoadProfileForName(name)
        if profile:
            components = self._ParseAddress(name)
            if not components["symbol"]:
                raise ValueError("No symbol name specified.")

            return profile.get_constant_object(
                components["symbol"],
                target=target, **kwargs)

        return obj.NoneObject("Profile for name %s unknown." % name, log=True)

    def _resolve_module_base_address(self, name):
        module = self.modules_by_name.get(name)
        if module is None:
            return obj.NoneObject("No module %s" % name, log=True)

        return module.base

    def get_address_by_name(self, name):
        self._EnsureInitialized()

        try:
            return int(name)
        except (ValueError, TypeError):
            pass

        if not isinstance(name, basestring):
            raise TypeError("Name should be a string.")

        components = self._ParseAddress(name)
        module_name = components["module"]

        address = components["address"]
        if address:
            address = int(address, 0)

        # User did not specify an address
        if address is None:
            # Found the module we use its base address
            address = self._resolve_module_base_address(module_name)
            if address == None:
                return obj.NoneObject(
                    "No module %s found" % module_name, log=True)

        # Search for a symbol in the module.
        if components["symbol"]:
            # Get the profile for this module.
            module_profile = self.LoadProfileForName(module_name)
            if module_profile:
                address = module_profile.get_constant(
                    components["symbol"], True)
            else:
                return obj.NoneObject(
                    "No profile found for module", log=True)

        # Support basic offset operations (+/-).
        if components["op"]:
            # Parse the offset as hex or decimal.
            offset = int(components["offset"], 0)
            op = components["op"].strip()
            if op == "+":
                address += offset
            elif op == "-":
                address -= offset
            else:
                raise TypeError("Operator '%s' not supported" % op)

        # If the symbol was a dereference, we need to read the address from
        # this offset.
        if components.get("deref"):
            module_profile = None
            containing_module = self._FindContainingModule(address)
            if containing_module:
                module_profile = self.LoadProfileForName(containing_module.name)

            if not module_profile:
                module_profile = self.session.profile

            address = module_profile.Pointer(address).v()

        return address

    def format_address(self, address, max_distance=0x1000):
        """Format the address as a symbol name.

        This means to try and find the containing module, the symbol within the
        module or possibly an offset from a known symbol. e.g.

        nt!PspCidTable
        nt!PspCidTable + 0x10
        nt + 0x234

        Returns an empty string if the address is not in a containing module, or
        if the nearest known symbol is farther than max_distance away.
        """
        _ = address
        _ = max_distance
        return ""

    def get_nearest_constant_by_address(self, address):
        """Searches for a known symbol at an address lower than this.

        Returns a tuple (nearest_offset, full_name of symbol).
        """
        _ = address
        return (0xFFFFFFFFFF, "")

    def search_symbol(self, pattern):
        """Searches symbols for the pattern.

        pattern may contain wild cards (*). Note that currently a module name is
        required. Example pattern:

        nt!Ps*
        """



# Rekall Memory Forensics
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

__author__ = "Michael Cohen <scudette@google.com>"


from rekall import testlib
from rekall.plugins.linux import common
from rekall.plugins.overlays import basic


class Netstat(common.LinuxPlugin):
    """Print the active network connections."""

    __name = "netstat"

    def sockets(self):
        """Enumerate all socket objects."""

        lsof = self.session.plugins.lsof(session=self.session)
        for task, file_struct, fd in lsof.lsof():
            if (file_struct.f_op == self.profile.get_constant(
                "socket_file_ops") or
                file_struct.m("d_entry").d_op == self.profile.get_constant(
                    "sockfs_dentry_operations")):

                iaddr = file_struct.dentry.d_inode

                # See http://lxr.free-electrons.com/source/include/net/sock.h?v=3.8#L1319
                skt = basic.container_of(iaddr, "socket_alloc",
                                         "vfs_inode").socket

                yield task, fd, skt.sk, iaddr

    def render(self, renderer):
        unix_sockets = []
        inet_sockets = []

        for task, fd, sock, iaddr in self.sockets():
            inet_sock = sock.dereference_as("inet_sock")

            if sock.sk_protocol not in ("IPPROTO_TCP", "IPPROTO_UDP",
                                        "IPPROTO_IPV4", "IPPROTO_IPV6",
                                        "IPPROTO_HOPOPT"):
                continue

            sk_common = sock.m("__sk_common")

            if sk_common.skc_family in ("AF_UNIX", "AF_LOCAL"):
                unix_sock = sock.dereference_as("unix_sock")
                name = unix_sock.addr.name[0].sun_path
                unix_sockets.append((task, fd, sock, iaddr, sk_common))

            elif sk_common.skc_family in ("AF_INET", "AF_INET6"):
                inet_sockets.append((task, fd, sock, iaddr, sk_common))

        # First do the AF_INET and AF_INET6 sockets.
        renderer.table_header([("Proto", "proto", "12"),
                               ("SAddr", "saddr", "15"),
                               ("SPort", "sport", "8"),
                               ("DAddr", "daddr", "15"),
                               ("DPort", "dport", "5"),
                               ("State", "state", "15"),
                               ("Pid", "pid", "8"),
                               ("Comm", "comm", "20")])

        for task, fd, sock, iaddr, sk_common in inet_sockets:
            inet_sock = sock.dereference_as("inet_sock")

            renderer.table_row(
                inet_sock.sk.sk_protocol,
                inet_sock.src_addr,
                inet_sock.src_port,
                inet_sock.dst_addr,
                inet_sock.dst_port,
                sk_common.skc_state,
                task.pid,
                task.comm,
                )

        renderer.section()

        # Now do the UNIX sockets.
        renderer.table_header([("Proto", "proto", "12"),
                               ("RefCount", "ref", "^8"),
                               ("Type", "type", "15"),
                               ("State", "state", "18"),
                               ("Inode", "inode", "8"),
                               ("Path", "path", "20")])

        for task, fd, sock, iaddr, sk_common in unix_sockets:
            unix_sock = sock.dereference_as("unix_sock")
            name = unix_sock.addr.name[0].sun_path

            renderer.table_row(
                "UNIX",
                unix_sock.addr.refcnt.counter,
                sock.sk_type,
                sk_common.skc_state,
                iaddr.i_ino,
                name
                )


class PacketQueues(common.LinuxPlugin):
    """Dumps the current packet queues for all known open sockets."""

    __name = "pkt_queues"

    @classmethod
    def args(cls, parser):
        super(PacketQueues, cls).args(parser)
        parser.add_argument(
            "--dump_dir", default=None, help="Output directory",
            required=True)

    def __init__(self, dump_dir=None, **kwargs):
        super(PacketQueues, self).__init__(**kwargs)
        self.output_dir = dump_dir

    def process_socket(self, renderer, task, fd_num, socket):
        queues = ["receive", "write"]
        for queue_name in queues:
            sk_buff_head_name = "sk_{0:s}_queue".format(queue_name)
            queue = getattr(socket, sk_buff_head_name)
            filename = "{0:d}.{1:s}.{2:s}.{3:d}".format(
                task.pid, task.name, queue_name, fd_num)
            data = []
            for sk_buff in queue.walk_list("next", False):
                pkt_len = sk_buff.len
                if pkt_len > 0:
                    if not sk_buff.data:
                        continue
                    data.append(self.kernel_address_space.read(
                        sk_buff.data.obj_offset, pkt_len))
            if data:
                with renderer.open(directory=self.output_dir,
                                   filename=filename,
                                   mode="wb") as fd:
                    fd.write(''.join(data))
                    self.session.logging.debug("Wrote %d bytes to %s",
                                               fd.tell(), filename)
                return True

            else:
                self.session.logging.debug("Skipped empty queue %s", filename)
                return False

    def render(self, renderer):
        netstat_plugin = self.session.plugins.netstat(session=self.session)
        sockets = netstat_plugin.sockets()

        all_sockets = list(sockets)
        skipped_sockets_count = 0
        for task, fd, socket, _ in all_sockets:
            if not self.process_socket(renderer, task, fd, socket):
                skipped_sockets_count += 1

        renderer.format("Skipped {0:d}/{1:d} sockets.", skipped_sockets_count,
                        len(all_sockets))


class TestPacketQueues(testlib.HashChecker):
    PARAMETERS = dict(
        commandline="pkt_queues --dump_dir %(tempdir)s"
        )



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
The Rekall Entity Layer.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

import copy
import itertools
import logging
import traceback

from rekall.entities import collector as entity_collector
from rekall.entities import component as entity_component
from rekall.entities import entity as entity_module
from rekall.entities import definitions
from rekall.entities import identity as entity_id
from rekall.entities import lookup_table as entity_lookup

from efilter import expression
from efilter import query as entity_query
from efilter.engines import matcher as query_matcher


class IngestionPipeline(object):
    """Keeps track of new and updated entities during collection."""

    empty = True

    def __init__(self, queries):
        self.queues = {}
        self.matchers = {}
        self.outcomes = {}

        for query in queries:
            self.queues[query] = []
            self.matchers[query] = query_matcher.ObjectMatcher(query)

    def seed(self, query, entities):
        """Set up the queue for query with entities."""
        self.queues[query] = list(entities)
        if self.queues[query]:
            self.empty = False

    def find(self, query):
        """Return entities available to satisfy the query."""
        if isinstance(query, dict):
            results = {}
            for key, value in query.iteritems():
                results[key] = self.find(value)
            return results

        return self.queues[query]

    def fill(self, ingest, collector, wanted_matcher=None, wanted_handler=None):
        """Fills appropriate queues with entities from ingest.

        Arguments:
            ingest: An iterable containing entities and effects of adding them.
                The effects are a dict of:
                    None: How many entities were duplicates, including contents.
                    entity_collector.EffectEnum.Merged: How many entities
                        changed by merging.
                    entity_collector.EffectEnum.Added: How many new entities
                        created.
                    "fill": How many entities were enqueued for ingestion by
                        other collectors.
            collector: The collector object from which ingest was collected.
        """

        counts = {entity_collector.EffectEnum.Duplicate: 0,
                  entity_collector.EffectEnum.Merged: 0,
                  entity_collector.EffectEnum.Added: 0,
                  entity_collector.EffectEnum.Enqueued: 0}

        for entity, effect in ingest:
            counts[effect] += 1

            if effect == entity_collector.EffectEnum.Duplicate:
                continue

            if wanted_handler and wanted_matcher.run(entity):
                wanted_handler(entity)

            for query in self.queries:
                if self.matchers[query].run(entity):
                    self.queues[query].append(entity)
                    counts[entity_collector.EffectEnum.Enqueued] += 1
                    self.empty = False

        if any(counts.itervalues()):
            logging.debug(
                "%s results: %d new, %d updated, %d requeued, %d duplicates.",
                collector.name,
                counts[entity_collector.EffectEnum.Added],
                counts[entity_collector.EffectEnum.Merged],
                counts[entity_collector.EffectEnum.Enqueued],
                counts[entity_collector.EffectEnum.Duplicate])

        self.outcomes[collector] = counts
        return counts

    def flush(self):
        queries = self.queues.keys()
        for query in queries:
            self.queues[query] = []
        self.empty = True

    def __getitem__(self, key):
        return self.queues[key]

    @property
    def queries(self):
        return set(self.queues.keys())


class EntityManager(object):
    """Database of entities."""

    # Names of collectors that have produced all they're going to produce.
    finished_collectors = None

    # Dict of entities keyed by their identity.
    entities = None

    def __init__(self, session):
        self.session = session
        self.reset()

    def reset(self):
        self.entities = {}
        self._collectors = {}
        self.finished_collectors = set()
        self._cached_query_analyses = {}
        self._cached_matchers = {}

        # Lookup table on component name is such a common use case that we
        # always have it on. This actually speeds up searches by attribute that
        # don't have a specific lookup table too.
        def _component_indexer(entity):
            for component in entity_component.Component.classes.keys():
                if getattr(entity.components, component):
                    yield component

        def _collector_indexer(entity):
            for collector_name in entity.components.Entity.collectors:
                yield collector_name

        self.lookup_tables = {
            "components": entity_lookup.EntityLookupTable(
                key_name="components",
                key_func=_component_indexer,
                entity_manager=self),
            "collectors": entity_lookup.EntityLookupTable(
                key_name="collectors",
                key_func=_collector_indexer,
                entity_manager=self)}

    @property
    def collectors(self):
        self.update_collectors()
        return self._collectors

    def update_collectors(self):
        """Refresh the list of active collectors. Do a diff if possible."""
        for key, cls in entity_collector.EntityCollector.classes.iteritems():
            if key in self._collectors:
                if cls.is_active(self.session):
                    continue
                else:
                    del self._collectors[key]
            else:
                if cls.is_active(self.session):
                    self._collectors[key] = cls(entity_manager=self)

    @property
    def identity_prefix(self):
        """Returns the prefix for all identities on this machine.

        Currently this just returns "LOCALHOST", but in the future this will
        be a way of semi-uniquelly identifying the image/machine of origin.
        """
        # TODO: Implement proper machine identification.
        return "LOCALHOST"

    def identify(self, identity_dict):
        """Generate the appropriate type of identity based on identity dict.

        Arguments:
            identity_dict: a dictionary of attribute names (format being the
            usual "Component/member") and expected values.

        Returns:
            An instance of Identity initialized with the identity dict and this
            manager's global prefix.
        """
        # Cast values to their correct types.
        cast_dict = {}
        for key, val in identity_dict.iteritems():
            if isinstance(key, tuple):
                cast_vals = []
                for idx, attr in enumerate(key):
                    attribute = entity_module.Entity.reflect_attribute(attr)
                    cast_vals.append(attribute.typedesc.coerce(val[idx]))
                cast_val = tuple(cast_vals)
            else:
                attribute = entity_module.Entity.reflect_attribute(key)
                cast_val = attribute.typedesc.coerce(val)

            cast_dict[key] = cast_val

        return entity_id.Identity.from_dict(global_prefix=self.identity_prefix,
                                            identity_dict=cast_dict)

    def identify_no_cast(self, identity_dict):
        return entity_id.Identity.from_dict(global_prefix=self.identity_prefix,
                                            identity_dict=identity_dict)

    # pylint: disable=protected-access
    def register_components(self, identity, components, source_collector):
        """Find or create an entity for identity and add components to it.

        Arguments:
            identity: What the components are about. Should be a subclass of
                Identity. As a special case, we also accept BaseObjects.

            components: An iterable of components about the identity.

            source_collector: Anything that responds to __unicode__ or __name__
                and describes the source of this information (usually the
                string name of the collector function).

        Returns:
            Tuple of entity and the effect of the new information.

            The effect can be one of:
                EffectEnum.Duplicate: No new information learned.
                EffectEnum.Merged: As result of this call, data in one or more
                    entities was updated and entities may have merged.
                EffectEnum.Added: A new entity was added.
        """

        kwargs = {}
        for component in components:
            kwargs[component.component_name] = component

        kwargs["Entity"] = definitions.Entity(
            identity=identity,
            collectors=frozenset((source_collector,)))

        entity = entity_module.Entity(
            entity_manager=self,
            components=entity_component.CONTAINER_PROTOTYPE._replace(**kwargs))

        indices = set(entity.indices)

        existing_entities = list(self.find_by_identity(identity))
        effect = entity_collector.EffectEnum.Added
        if existing_entities:
            if (len(existing_entities) == 1 and
                    existing_entities[0].issuperset(entity)):
                # No new data, but let's give the collector credit for finding
                # what we already knew.
                entity_comp = existing_entities[0].components.Entity
                entity_comp._mutate(
                    member="collectors",
                    value=entity_comp.collectors.union([source_collector]))

                return (existing_entities[0],
                        entity_collector.EffectEnum.Duplicate)

            for existing_entity in existing_entities:
                # Entities exist for this already, but are not equivalent to
                # the entity we found. Merge everything.
                effect = entity_collector.EffectEnum.Merged
                entity.update(existing_entity)
                indices.update(existing_entity.indices)

        # Overwrite all old indices with reference to the new entity.
        for index in indices:
            self.entities[index] = entity

        for lookup_table in self.lookup_tables.itervalues():
            lookup_table.update_index((entity,))

        return entity, effect

    def add_attribute_lookup(self, key):
        """Adds a fast-lookup index for the component/attribute key path.

        This also causes the newly-created lookup table to rebuild its index.
        Depending on how many entities already exist, building the index could
        take a couple of seconds.
        """
        # Don't add the same one twice.
        if self.lookup_tables.get(key, None):
            return

        attribute = entity_module.Entity.reflect_attribute(key)
        if not isinstance(attribute, entity_component.Field):
            logging.info(
                ("Can't create a lookup for %s, because it's not a simple "
                 "field."), attribute)
            return

        logging.debug("Creating a lookup table for %s", key)
        component, _ = key.split("/")

        lookup_table = entity_lookup.AttributeLookupTable(
            attribute=key,
            entity_manager=self)

        # Only use the entities that actually have the component to build the
        # index.
        lookup_table.update_index(
            self.find(expression.ComponentLiteral(component), complete=False,
                      keep_cache=True))

        self.lookup_tables[key] = lookup_table

    def find_by_identity(self, identity, complete=False):
        """Yield the entities that matches the identity.

        The number of entities yielded is almost always one or zero. The single
        exception to that rule is when the identity parameter is both: (a) a
        alternate identity and (b) not yet present in this entity manager. In
        that case, multiple entities may match.

        Arguments:
            identity: The identity to search for.
            complete: Should collectors be run to ensure complete results?
        """
        if complete:
            self.collect_for(identity.as_query())

        results = set()
        for index in identity.indices:
            entity = self.entities.get(index, None)
            if entity:
                results.add(entity)

        if complete:
            results = [self.parse(entity) for entity in results]

        return list(results)

    def find_by_component(self, component, complete=True):
        """Finds all entities that have the component.

        Arguments:
            complete: If True, will attempt to collect the component.
        """
        query = entity_query.Query(expression.ComponentLiteral(component))
        if complete:
            self.collect_for(query)

        return list(self.lookup_tables["components"].lookup(component))

    def find_by_collector(self, collector):
        """Find all entities touched by the collector."""
        return list(self.lookup_tables["collectors"].lookup(str(collector)))

    def matcher_for(self, query):
        """Returns a query matcher for the query (cached)."""
        matcher = self._cached_matchers.setdefault(
            query, query_matcher.ObjectMatcher(query))

        return matcher

    def parsers_for(self, entity):
        """Finds collectors that can parse this entity.

        Yields: tuples of:
            - collector instance
            - name of the keyword argument on the collect method under which
              the entity should be passed to the collector.
        """
        for collector in self.collectors.itervalues():
            if len(collector.collect_queries) != 1:
                continue
            for query_name, query in collector.collect_queries.iteritems():
                matcher = self.matcher_for(query)
                if matcher.run(entity):
                    yield collector, query_name

    def parse(self, entity):
        """Parses the entity using available higher-order collectors."""
        result = entity
        for collector, collect_kwarg in self.parsers_for(entity):
            collector_input = {collect_kwarg: [result]}
            for parsed, effect in self.collect(collector=collector,
                                               collector_input=collector_input,
                                               hint=None):
                if effect != entity_collector.EffectEnum.Duplicate:
                    logging.debug(
                        "Collector %s produced a hit in parser mode.",
                        collector.name)
                    result = parsed

        return result

    def analyze(self, wanted):
        """Finds collectors and indexing suggestions for the query.

        Returns a dict of:
            - collectors: list of collectors to run
            - lookups: list of attributes to consider indexing for
            - dependencies: list of SimpleDependency instances to include
            - exclusions: list of SimpleDependency instances to exclude
        """
        if not isinstance(wanted, entity_query.Query):
            wanted = entity_query.Query(wanted)

        # We cache by the source and not the query because we want to reanalyze
        # queries that are logically equivalent, but expressed differently, in
        # order to have the right cursor positions stored for highlighting in
        # GUI.
        cache_key = wanted.source

        analysis = self._cached_query_analyses.get(cache_key, None)
        if analysis:
            # We want to make a copy exactly one level deep.
            analysis_copy = {}
            for key, value in analysis.iteritems():
                analysis_copy[key] = copy.copy(value)
            return analysis_copy

        analyzer = wanted.run_engine("slashy_analyzer")

        include = set()
        for dependency in analyzer.include:
            # Skip over general dependencies on Entity.
            if dependency.component == "Entity" and not dependency.attribute:
                continue

            include.add(dependency)

        exclude = analyzer.exclude
        suggested_indices = analyzer.latest_indices

        # A collector is a match if any of its promises match any of the
        # dependencies of the query.
        matched_collectors = []
        for collector in self.collectors.itervalues():
            for promise, dependency in itertools.product(
                    collector.promises, include):
                if dependency.match(promise):
                    matched_collectors.append(collector)
                    break

        # A collector is yielded unless each one of its promises matches
        # an exclusion from dependencies.
        collectors = set()
        for collector in matched_collectors:
            for promise, exclusion in itertools.product(
                    collector.promises, exclude):
                if not exclusion.match(promise):
                    collectors.add(collector)
                    break
            else:
                # No exclusions.
                collectors.add(collector)

        # A component is guaranteed if any dependency lists it. It is likely
        # if collectors we depend on output it (though not guaranteed).
        guaranteed_components = set(analyzer.expected_components)
        possible_components = set()

        for dependency in include:
            component = dependency.component
            if component in guaranteed_components:
                continue
            possible_components.add(dependency.component)

        for collector in collectors:
            for promise in collector.promises:
                component = promise.component
                if component in guaranteed_components:
                    continue

                possible_components.add(component)

        analysis = dict(collectors=list(collectors),
                        lookups=suggested_indices,
                        dependencies=include,
                        exclusions=exclude,
                        guaranteed_components=guaranteed_components,
                        possible_components=possible_components)
        self._cached_query_analyses[cache_key] = analysis

        return analysis

    def find(self, query, complete=True, validate=True, query_params=None,
             retry_on_error=False, keep_cache=False, syntax="slashy"):
        """Runs the query and yields entities that match.

        Arguments:
            query: Either an instance of the query AST, a query string, or a
                   dictionary of queries. If a dict is given, a new dict will
                   be returned with the same keys and values replaced with
                   results.

            query_params: If query accepts parameters (it's a template), you
                          may pass them here.

            complete: If True, will trigger collectors as necessary, to ensure
                      completness of results.

            validate: Will cause the query to be validated first (mostly for
                      type errors).

        Arguments (live analysis only):
            retry_on_error: Should query be retried on failed collection?

                            This argument is only respected on live systems.

            keep_cache: Should we reuse cached data from previous searches?
                        This will greatly speed up analysis, but may lead to
                        outdated or inconsistent results on running systems.

                        This argument is only respected on live systems.
        """
        if not self.session.volatile:
            keep_cache = True
            retry_on_error = False

        if not keep_cache:
            if not complete:
                raise ValueError(
                    "keep_cache and complete cannot both be False.")

            self.reset()

        if isinstance(query, dict):
            results = {}
            for query_name, expr in query.iteritems():
                results[query_name] = self.find(expr, complete=complete,
                                                validate=validate,
                                                query_params=query_params,
                                                retry_on_error=retry_on_error,
                                                keep_cache=keep_cache)

            return results

        if not isinstance(query, entity_query.Query):
            query = entity_query.Query(query, params=query_params,
                                       syntax=syntax)

        if validate:
            query.run_engine("validator")

        if complete:
            try:
                self.collect_for(query)
            except (entity_id.IdentityError, TypeError) as e:
                if retry_on_error:
                    logging.error(
                        "Collect failed for query %r. It would appear Rekall "
                        "is running on live memory - it is possible contents "
                        "of memory changed between reads, causing a collector "
                        "to fail. Going to try again. Original error: %r\n%s",
                        query, e, traceback.format_exc())

                    self.reset()
                    return self.find(query=query, complete=True,
                                     validate=validate,
                                     query_params=query_params,
                                     keep_cache=keep_cache,
                                     retry_on_error=False)
                else:
                    # We're not retrying and/or running on an image. Just
                    # rethrow the exception here.
                    raise

        # Try to satisfy the query using available lookup tables.
        search = entity_lookup.EntityQuerySearch(query)
        return search.search(self.entities, self.lookup_tables)

    def stream(self, query, handler, query_params=None):
        query = entity_query.Query(query, params=query_params)
        seen = set()

        def _deduplicator(entity):
            if entity in seen:
                return

            seen.add(entity)
            handler(entity)

        self.collect_for(query, result_stream_handler=_deduplicator)
        for entity in self.find(query, complete=False, keep_cache=True):
            _deduplicator(entity)

    def find_first(self, query, complete=True, validate=True,
                   query_params=None):
        """Like find, but returns just the first result."""
        for entity in self.find(query, complete, validate, query_params):
            return entity

    # pylint: disable=protected-access
    def collect_for(self, wanted, use_hint=False, result_stream_handler=None):
        """Will find and run the appropriate collectors to satisfy the query.

        If use_hint is set to True, 'wanted' will be passed on as hint to
        the collectors. This may result in faster collection, but may result
        in collectors having to run repeatedly.
        """
        # Planning stage.

        if callable(result_stream_handler):
            wanted_matcher = query_matcher.ObjectMatcher(wanted)
        else:
            wanted_matcher = None

        self.update_collectors()

        # to_process is used as a FIFO queue below.
        analysis = self.analyze(wanted)
        to_process = analysis["collectors"][:]
        suggested_indices = analysis["lookups"]

        # Create indices as suggested by the analyzer.
        for attribute in suggested_indices:
            self.add_attribute_lookup(attribute)

        collectors_seen = set(self.finished_collectors)

        # Collectors with an ingest query are de-facto parsers for things
        # produced by collectors with no ingest query. They may run repeatedly
        # as required.
        repeated = list()

        # Collectors with no dependencies (my favorite).
        simple = list()

        # Queries that collectors depend on.
        queries = set()

        # Build up a list of collectors to run, based on dependencies.
        while to_process:
            collector = to_process.pop(0)
            if collector.name in collectors_seen:
                continue

            collectors_seen.add(collector.name)
            if collector.collect_queries:
                logging.debug("Collector %s deferred until stage 2.",
                              collector.name)
                repeated.append(collector)
                queries |= set(collector.collect_queries.itervalues())

                # Discard the indexing suggestions for ingestion queries
                # because they don't represent normal usage.
                additional = set()
                for query in collector.collect_queries.itervalues():
                    additional |= set(self.analyze(query)["collectors"])

                for dependency in additional:
                    logging.debug("Collector %s depends on collector %s.",
                                  collector.name, dependency.name)
                    if dependency.name not in collectors_seen:
                        to_process.append(dependency)
            else:
                logging.debug("%s will run in stage 1.",
                              collector.name)
                simple.append(collector)

        if not collectors_seen.difference(self.finished_collectors):
            # Looks like we're already populated - no need to do anything.
            return

        logging.info(
            "Will now run %d first-order collectors and %d collectors with "
            "dependencies to satisfy query %s.",
            len(simple), len(repeated), wanted)

        # Execution stage 1: no dependencies.

        for collector in simple:
            effects = {entity_collector.EffectEnum.Duplicate: 0,
                       entity_collector.EffectEnum.Merged: 0,
                       entity_collector.EffectEnum.Added: 0}

            if use_hint or collector.enforce_hint:
                hint = wanted
            else:
                hint = None
                self.finished_collectors.add(collector.name)

            for entity, effect in self.collect(collector, hint=hint):
                if result_stream_handler and wanted_matcher.run(entity):
                    result_stream_handler(entity)

                effects[effect] += 1

            logging.debug(
                "%s produced %d new entities, %d updated and %d duplicates",
                collector.name,
                effects[entity_collector.EffectEnum.Added],
                effects[entity_collector.EffectEnum.Merged],
                effects[entity_collector.EffectEnum.Duplicate])

        if not repeated:
            # No higher-order collectors scheduled. We're done.
            return

        # Seeding stage for higher-order collectors.
        in_pipeline = IngestionPipeline(queries=queries)
        out_pipeline = IngestionPipeline(queries=queries)
        for query in queries:
            results = self.find(query, complete=False, keep_cache=True)
            in_pipeline.seed(query, results)
            if results:
                logging.debug("Pipeline seeded with %d entities matching '%s'",
                              len(results), query)

        # Execution stage 2: collectors with dependencies.

        # Collectors should run in FIFO order:
        repeated.reverse()

        repeat_counter = 0
        # This will spin until none of the remaining collectors want to run.
        while not in_pipeline.empty:
            # Collectors will read from the in_pipeline and fill the
            # out_pipeline. At the end of each spin the pipelines swap and
            # the new out_pipeline is flushed.
            for collector in repeated:
                # If the collector wants complete input, we pull it from the
                # database. If it just wants one entity at a time, we can use
                # the ingestion pipeline. The semantics of both find methods
                # are identical.
                if collector.complete_input:
                    collector_input = self.find(collector.collect_queries,
                                                complete=False,
                                                keep_cache=True)
                else:
                    collector_input = in_pipeline.find(
                        collector.collect_queries)

                # The collector requests its prefilter to be called.
                if collector.filter_input:
                    collector_input_filtered = {}
                    for key, val in collector_input.iteritems():
                        collector_input_filtered[key] = collector.input_filter(
                            hint=hint, entities=val)
                    collector_input = collector_input_filtered

                # The collector requests that we always pass the query hint.
                if use_hint or collector.enforce_hint:
                    hint = wanted
                else:
                    hint = None

                try:
                    # Feed output back into the pipeline.
                    results = self.collect(collector=collector,
                                           collector_input=collector_input,
                                           hint=hint)

                    out_pipeline.fill(collector=collector,
                                      ingest=results,
                                      wanted_handler=result_stream_handler,
                                      wanted_matcher=wanted_matcher)
                except entity_id.IdentityError as e:
                    logging.error(
                        "Collector %r has encountered invalid or inconsistent "
                        "data and could not recover. Details available with "
                        "debug logging." % collector)
                    logging.debug(
                        "Collector %r (hint %r) raised %r\n%s.\n"
                        "Collector input was %r."
                        % (collector, e, traceback.format_exc(),
                           collector_input, hint))

            # Check for endless loops.
            if in_pipeline.outcomes == out_pipeline.outcomes:
                repeat_counter += 1
                if repeat_counter < 5:
                    logging.warning(
                        "Detected a loop in collection run (%d cycles)." %
                        repeat_counter)
                else:
                    logging.warning(
                        "Maximum number of cycles in collection run exceeded. "
                        "Terminating collection.")
                    break
            else:
                repeat_counter = 0

            # Swap & flush, rinse & repeat.
            in_pipeline, out_pipeline = out_pipeline, in_pipeline
            out_pipeline.flush()

        for collector in repeated:
            if not use_hint and not collector.enforce_hint:
                self.finished_collectors.add(collector.name)

    def collect(self, collector, hint, collector_input=None):
        """Runs the collector, registers output and yields any new entities."""
        if collector_input is None:
            collector_input = {}

        result_counter = 0

        if self.session:
            self.session.report_progress(
                "Collecting %(collector)s %(spinner)s",
                collector=collector.name)

        for results in collector.collect(hint=hint, **collector_input):
            if not isinstance(results, list):
                # Just one component yielded.
                results = [results]

            # First result is either the first component or an identity.
            first_result = results[0]
            if isinstance(first_result, entity_id.Identity):
                # If the collector gave as an identity then use that.
                identity = first_result
                results.pop(0)
            else:
                # If collector didn't give us an identity then we build
                # one from the first component's first field. This is
                # a good heuristic for about 90% of the time.
                first_field = first_result.component_fields[0].name
                attribute = "%s/%s" % (type(first_result).__name__,
                                       first_field)
                try:
                    identity = self.identify({attribute: first_result[0]})
                except entity_id.IdentityError:
                    logging.warning(
                        ("Invalid identity %r inferred from output of %r. "
                         "Entity skipped. Full results: %r"),
                        {attribute: first_result[0]},
                        collector,
                        results)
                    continue

            try:
                entity, effect = self.register_components(
                    identity=identity,
                    components=results,
                    source_collector=collector.name)
            except entity_id.IdentityError as e:
                logging.warning(
                    ("Invalid identity %r inferred from output of %r. "
                     "Entity skipped. Full results: %r. "
                     "Original error: %s"),
                    identity, collector, results, e)
                continue

            result_counter += 1
            if result_counter % 100 == 0 and self.session:
                self.session.report_progress(
                    "Collecting %(collector)s %(spinner)s (%(count)d results)",
                    collector=collector.name,
                    count=result_counter)

            yield entity, effect



# Rekall Memory Forensics
# Copyright (C) 2007,2008 Volatile Systems
# Copyright (C) 2010,2011,2012 Michael Hale Ligh <michael.ligh@mnin.org>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""Most of the following structures are actually documented in Windows 7
onwards, but are not documented in windows XP.

For those structs which did not change between Windows XP and Windows 7, we can
just copy the ones from Win7 but some things have changes so we still need to
hard code the following.

Ideally we should use generate_types here.
"""

from rekall.plugins.windows.gui import constants


vtypes_xp_32 = {
    'tagWINDOWSTATION' : [0x5C, {
        'dwSessionId' : [0x0, ['unsigned long']],
        'rpwinstaNext' : [0x4, ['pointer', ['tagWINDOWSTATION']]],
        'rpdeskList' : [0x8, ['pointer', ['tagDESKTOP']]],
        'dwWSF_Flags' : [0x10, ['unsigned long']],
        'ptiDrawingClipboard' : [0x1C, ['pointer', ['tagTHREADINFO']]],
        'spwndClipOpen' : [0x20, ['pointer', ['tagWND']]],
        'spwndClipViewer' : [0x24, ['pointer', ['tagWND']]],
        'spwndClipOwner' : [0x28, ['pointer', ['tagWND']]],
        'pClipBase' : [0x2C, ['pointer', ['array', lambda x : x.cNumClipFormats, ['tagCLIP']]]],
        'cNumClipFormats' : [0x30, ['unsigned int']],
        'iClipSerialNumber' : [0x34, ['unsigned int']],
        'iClipSequenceNumber' : [0x38, ['unsigned int']],
        #'spwndClipboardListener' : [0x3C, ['pointer', ['tagWND']]],
        'pGlobalAtomTable' : [0x40, ['pointer', ['void']]],
        }],

    'tagDESKTOP' : [0x84, {
        'dwSessionId' : [0x0, ['unsigned long']],
        'pDeskInfo' : [0x4, ['pointer', ['tagDESKTOPINFO']]],
        'rpdeskNext' : [0xc, ['pointer', ['tagDESKTOP']]],
        'rpwinstaParent' : [0x10, ['pointer', ['tagWINDOWSTATION']]],
        'hsectionDesktop' : [0x40, ['pointer', ['void']]],
        'pheapDesktop' : [0x44, ['pointer', ['tagWIN32HEAP']]],
        'PtiList' : [0x64, ['_LIST_ENTRY']],
        }],

    'tagTHREADINFO' : [None, {
        'pEThread' : [0x00, ['pointer', ['_ETHREAD']]],
        'ppi' : [0x2C, ['pointer', ['tagPROCESSINFO']]],
        'pq' : [0x30, ['pointer', ['tagQ']]],
        'pDeskInfo' : [0x40, ['pointer', ['tagDESKTOPINFO']]],
        'PtiLink' : [0xAC, ['_LIST_ENTRY']],
        'fsHooks' : [0x98, ['unsigned long']],
        'aphkStart' : [0xF4, ['array', 16, ['pointer', ['tagHOOK']]]],
        }],

    'tagQ' : [None, {
        'mlInput' : [0x00, ['tagMLIST']],
        }],

    'tagMLIST' : [None, {
        'pqmsgRead' : [0x00, ['pointer', ['tagQMSG']]],
        'cMsgs' : [0x08, ['unsigned long']],
        }],
    'tagQMSG' : [None, {
        'pqmsgNext' : [0x00, ['pointer', ['tagQMSG']]],
        'pqmsgPrev' : [0x04, ['pointer', ['tagQMSG']]],
        'msg' : [0x08, ['tagMSG']],
        }],
    'tagMSG' : [None, {
        'hwnd' : [0x00, ['unsigned long']],
        'message' : [0x04, ['unsigned long']],
        'wParam' : [0x08, ['unsigned long']],
        'lParam' : [0x0C, ['unsigned long']],
        'time' : [0x10, ['unsigned long']],
        'pt' : [0x14, ['tagPOINT']],
        }],
    'tagPOINT' : [None, {
        'x' : [0x00, ['long']],
        'y' : [0x04, ['long']],
        }],
    'tagHOOK' : [None, {
        'head' : [0x0, ['_THRDESKHEAD']],
        'phkNext' : [0x14, ['pointer', ['tagHOOK']]],
        'iHook' : [0x18, ['long']],
        'offPfn' : [0x1c, ['unsigned long']],
        'flags': [0x20, ['Flags', {'bitmap': constants.HOOK_FLAGS}]],
        'ihmod' : [0x24, ['long']],
        'ptiHooked' : [0x28, ['pointer', ['tagTHREADINFO']]],
        'rpdesk' : [0x2c, ['pointer', ['tagDESKTOP']]],
        }],
    'tagDESKTOPINFO' : [None, {
        'pvDesktopBase' : [0x0, ['pointer', ['void']]],
        'pvDesktopLimit' : [0x4, ['pointer', ['void']]],
        'spwnd' : [0x08, ['pointer', ['tagWND']]],
        'fsHooks' : [0x0c, ['unsigned long']],
        'aphkStart' : [0x10, ['array', 16, ['pointer', ['tagHOOK']]]],
        }],
    'tagSERVERINFO' : [0xffc, {
        'cHandleEntries' : [8, ['unsigned long']],
        'cbHandleTable' : [0x1bc, ['unsigned long']],
        }],

    'tagPROCESSINFO' : [None, {
        'Process' : [0x0, ['pointer', ['_EPROCESS']]],
        }],
    '_THRDESKHEAD' : [0x14, {
        'h' : [0x0, ['pointer', ['void']]],
        'cLockObj' : [0x4, ['unsigned long']],
        'pti' : [0x8, ['pointer', ['tagTHREADINFO']]],
        'rpdesk' : [0xc, ['pointer', ['tagDESKTOP']]],
        'pSelf' : [0x10, ['pointer', ['unsigned char']]],
        }],
    'tagCLS' : [0x5c, {
        'pclsNext' : [0x0, ['pointer', ['tagCLS']]],
        'atomClassName' : [0x4, ['unsigned short']],
        'atomNVClassName' : [0x6, ['unsigned short']],
        }],
    'tagRECT' : [0x10, {
        'left' : [0x0, ['long']],
        'top' : [0x4, ['long']],
        'right' : [0x8, ['long']],
        'bottom' : [0xc, ['long']],
        }],
    'tagWND' : [0x90, {
        'head' : [0x0, ['_THRDESKHEAD']],
        'ExStyle' : [0x1c, ['unsigned long']],
        'style' : [0x20, ['unsigned long']],
        'hModule' : [0x24, ['pointer', ['void']]],
        'spwndNext' : [0x2c, ['pointer', ['tagWND']]],
        'spwndPrev' : [0x30, ['pointer', ['tagWND']]],
        'spwndParent' : [0x34, ['pointer', ['tagWND']]],
        'spwndChild' : [0x38, ['pointer', ['tagWND']]],
        'spwndOwner' : [0x3c, ['pointer', ['tagWND']]],
        'rcWindow' : [0x40, ['tagRECT']],
        'rcClient' : [0x50, ['tagRECT']],
        'lpfnWndProc' : [0x60, ['pointer', ['void']]],
        'pcls' : [0x64, ['pointer', ['tagCLS']]],
        'strName' : [0x80, ['_LARGE_UNICODE_STRING']],
        'cbwndExtra' : [0x8C, ['long']],
        'dwUserData' : [0x98, ['unsigned long']],
        }],
    '_LARGE_UNICODE_STRING' : [0xc, {
        'Length' : [0x0, ['unsigned long']],
        'MaximumLength' : [0x4, ['BitField', dict(start_bit = 0, end_bit = 31)]],
        'bAnsi' : [0x4, ['BitField', dict(start_bit = 31, end_bit = 32)]],
        'Buffer' : [0x8, ['pointer', ['unsigned short']]],
        }],
    }

vtypes_xp_64 = {
    'tagWINDOWSTATION' : [0x90, { # !poolfind Wind is 100h
    'dwSessionId' : [0x0, ['unsigned long']],
    'rpwinstaNext' : [0x8, ['pointer64', ['tagWINDOWSTATION']]], # FreeWindowStation
    'rpdeskList' : [0x10, ['pointer64', ['tagDESKTOP']]],
    'dwWSF_Flags' : [0x20, ['unsigned long']], # FreeWindowStation
    'ptiDrawingClipboard' : [0x38, ['pointer64', ['tagTHREADINFO']]], # xxxDrawClipboard
    'spwndClipOpen' : [0x40, ['pointer64', ['tagWND']]],
    'spwndClipViewer' : [0x48, ['pointer64', ['tagWND']]],
    'spwndClipOwner' : [0x50, ['pointer64', ['tagWND']]],
    'pClipBase' : [0x58, ['pointer64', ['array', lambda x : x.cNumClipFormats, ['tagCLIP']]]], # InternalSetClipboardData
    'cNumClipFormats' : [0x60, ['unsigned int']], # InternalSetClipboardData
    'iClipSerialNumber' : [0x64, ['unsigned int']], # InternalSetClipboardData
    'iClipSequenceNumber' : [0x68, ['unsigned int']], # InternalSetClipboardData
    'pGlobalAtomTable' : [0x70, ['pointer64', ['void']]],
    }],

    # !poolfind Desk is 140h
    'tagDESKTOP' : [0xd0, {
        'dwSessionId' : [0x0, ['unsigned long']],
        'pDeskInfo' : [0x8, ['pointer64', ['tagDESKTOPINFO']]], # xxxCreateDesktop
        'rpdeskNext' : [0x18, ['pointer64', ['tagDESKTOP']]], # ParseDesktop
        'rpwinstaParent' : [0x20, ['pointer64', ['tagWINDOWSTATION']]],
        'hsectionDesktop' : [0x70, ['pointer64', ['void']]], # MapDesktop
        'pheapDesktop' : [0x78, ['pointer64', ['tagWIN32HEAP']]], # DesktopAlloc
        'PtiList' : [0xa0, ['_LIST_ENTRY']], # zzzJournalAttach
        }],

    'tagTHREADINFO' : [None, {
        'pEThread' : [0x00, ['pointer', ['_ETHREAD']]],
        'ppi' : [0x68, ['pointer64', ['tagPROCESSINFO']]], # xxxSetThreadDesktop
        #'pq' : [0x30, ['pointer', ['tagQ']]],
        'pDeskInfo' : [0x90, ['pointer64', ['tagDESKTOPINFO']]], # xxxDesktopThread
        'PtiLink' : [0x160, ['_LIST_ENTRY']],
        'fsHooks' : [0x138, ['unsigned long']], # xxxSetThreadDesktop, CheckWHFBits
        'aphkStart' : [0x140, ['array', 16, ['pointer64', ['tagHOOK']]]],
        }],

    'tagDESKTOPINFO' : [None, {
        'pvDesktopBase' : [0x0, ['pointer64', ['void']]],
        'pvDesktopLimit' : [0x8, ['pointer64', ['void']]],
        'spwnd' : [0x10, ['pointer64', ['tagWND']]],
        'fsHooks' : [0x18, ['unsigned long']], # CheckWHFBits
        'aphkStart' : [0x20, ['array', 16, ['pointer64', ['tagHOOK']]]],
        }],

    'tagWND' : [None, {
        'head' : [0x0, ['_THRDESKHEAD']],
        'ExStyle' : [0x30, ['unsigned long']], # xxxCreateWindowEx
        'style' : [0x34, ['unsigned long']], # xxxCreateWindowEx
        'spwndNext' : [0x48, ['pointer64', ['tagWND']]],
        'spwndPrev' : [0x50, ['pointer64', ['tagWND']]],
        'spwndParent' : [0x58, ['pointer64', ['tagWND']]],
        'spwndChild' : [0x60, ['pointer64', ['tagWND']]],
        'spwndOwner' : [0x68, ['pointer64', ['tagWND']]],
        'rcWindow' : [0x70, ['tagRECT']],
        'rcClient' : [0x80, ['tagRECT']],
        'lpfnWndProc' : [0x90, ['pointer64', ['void']]],
        'pcls' : [0x98, ['pointer64', ['tagCLS']]], # HMChangeOwnerThread
        'strName' : [0xd0, ['_LARGE_UNICODE_STRING']],
        }],

    'tagRECT' : [0x10, {
        'left' : [0x0, ['long']],
        'top' : [0x4, ['long']],
        'right' : [0x8, ['long']],
        'bottom' : [0xc, ['long']],
        }],

    'tagCLS' : [None, {
        'pclsNext' : [0x0, ['pointer64', ['tagCLS']]],
        'atomClassName' : [0x8, ['unsigned short']], # HMChangeOwnerThread
        'atomNVClassName' : [0xA, ['unsigned short']],
        }],
    }



# Rekall Memory Forensics
# Copyright (c) 2012, Michael Cohen <scudette@gmail.com>
# Copyright (c) 2010, 2011, 2012 Michael Ligh <michael.ligh@mnin.org>
# Copyright 2013 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

import yara

from rekall import plugin
from rekall import testlib
from rekall import utils

from rekall.plugins import yarascanner
from rekall.plugins.windows import common
from rekall.plugins.windows import vadinfo


class VadYaraScanner(vadinfo.VadScanner, yarascanner.BaseYaraASScanner):
    """A Yara scanner which only operates on VAD regions."""


class WinYaraScan(common.WinProcessFilter):
    """Scan using yara signatures."""

    name = "yarascan"

    @classmethod
    def args(cls, parser):
        super(WinYaraScan, cls).args(parser)
        parser.add_argument("--string", default=None,
                            help="A verbatim string to search for.")

        parser.add_argument("--binary_string", default=None,
                            help="A binary string (encoded as hex) to search "
                            "for. e.g. 000102[1-200]0506")

        parser.add_argument("--yara_file", default=None,
                            help="The yara signature file to read.")

        parser.add_argument("--yara_expression", default=None,
                            help="If provided we scan for this yarra "
                            "expression.")

        parser.add_argument(
            "--scan_vads", default=False, type="Boolean",
            help=("If specified we scan the vads of "
                  "processes, else we scan their entire address spaces. "
                  "Note that scanning the entire address space will "
                  "typically also include the kernel, while targetting "
                  "VADs only will only include the addresses mapped by "
                  "the process."))

        parser.add_argument(
            "--scan_physical", default=False, type="Boolean",
            help="If specified we scan the physcial address space.")

    def __init__(self, string=None, scan_vads=False, scan_physical=False,
                 yara_file=None, yara_expression=None, binary_string=None,
                 **kwargs):
        """Scan using yara signatures."""
        super(WinYaraScan, self).__init__(**kwargs)
        if yara_expression:
            self.rules_source = yara_expression
            self.rules = yara.compile(source=self.rules_source)

        elif binary_string:
            self.compile_rule(
                'rule r1 {strings: $a = {%s} condition: $a}' % binary_string
                )
        elif string:
            self.compile_rule(
                'rule r1 {strings: $a = "%s" condition: $a}' % string
                )

        elif yara_file:
            self.compile_rule(open(yara_file).read())
        else:
            raise plugin.PluginError("You must specify a yara rule file or "
                                     "string to match.")

        self.scan_vads = scan_vads
        self.scan_physical = scan_physical

    def compile_rule(self, rule):
        self.rules_source = rule
        try:
            self.rules = yara.compile(source=rule)
        except Exception as e:
            raise plugin.PluginError(
                "Failed to compile yara expression: %s" % e)

    def generate_hits(self, address_space, end=None):
        scanner = yarascanner.BaseYaraASScanner(
            profile=self.profile, session=self.session,
            address_space=address_space,
            rules=self.rules)

        return scanner.scan(maxlen=end)

    def render_scan_physical(self, renderer):
        """This method scans the physical memory."""
        for rule, address, _, _ in self.generate_hits(
                self.physical_address_space):
            renderer.format("Rule: {0}\n", rule)

            context = self.physical_address_space.read(address, 0x40)
            utils.WriteHexdump(renderer, context, base=address)

    def render_kernel_scan(self, renderer):
        modules = self.session.plugins.modules()

        for rule, address, _, _ in self.generate_hits(
                self.kernel_address_space):
            renderer.format("Rule: {0}\n", rule)

            # Find out who owns this hit.
            owner = modules.find_module(address)
            if owner:
                renderer.format("Owner: {0}\n", owner.BaseDllName)
            else:
                renderer.format("Owner: (Unknown Kernel Memory)\n")

            context = self.kernel_address_space.read(address, 0x40)
            utils.WriteHexdump(renderer, context, base=address)

    def render_task_scan(self, renderer, task):
        task_as = task.get_process_address_space()

        for rule, address, _, _ in self.generate_hits(task_as, end=end):
            renderer.format("Rule: {0}\n", rule)

            renderer.format("Owner: {0}\n", task.ImageFileName)

            context = task_as.read(address, 0x40)
            utils.WriteHexdump(renderer, context, base=address)

    def render_task_scan_vad(self, renderer, task):
        """This method scans the process memory using the VAD."""
        task_as = task.get_process_address_space()
        end = self.session.GetParameter("highest_usermode_address")

        scanner = VadYaraScanner(
            session=self.session, rules=self.rules, task=task)

        for rule, address, _, _ in scanner.scan(maxlen=end):
            renderer.format("Rule: {0}\n", rule)

            renderer.format("Owner: {0}\n", task.ImageFileName)

            context = task_as.read(address, 0x40)
            utils.WriteHexdump(renderer, context, base=address)

    def render(self, renderer):
        """Render output."""
        if self.scan_physical:
            return self.render_scan_physical(renderer)

        elif self.scan_vads or self.filtering_requested:
            for task in self.filter_processes():
                if self.scan_vads:
                    self.render_task_scan_vad(renderer, task)
                else:
                    self.render_task_scan(renderer, task)

        # We are searching the kernel address space
        else:
            return self.render_kernel_scan(renderer)


class TestYara(testlib.SimpleTestCase):
    """Test the yarascan module."""

    PARAMETERS = dict(commandline="yarascan --string sys")



#!/usr/bin/python

# Rekall
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""This file implements a caching URL manager.

We locally cache selected files from the remote profile repository. If a profile
is not found in the local cache, we retrieve it from the remote repository, and
add it to the cache.

We check the remote repository for staleness and update the local cache in the
background.
"""

__author__ = "Michael Cohen <scudette@google.com>"
import os

from rekall import config
from rekall import io_manager


config.DeclareOption(
    "cache_dir", default=None,
    help="Location of the profile cache directory.")


class CachingManager(io_manager.IOManager):

    # We wrap this io manager class
    DELEGATE = io_manager.URLManager

    # If the cache is available we should be selected before the regular
    # manager.
    order = DELEGATE.order - 10

    def __init__(self, session=None, **kwargs):
        cache_dir = session.GetParameter("cache_dir")

        if not cache_dir:
            raise io_manager.IOManagerError(
                "Local profile cache is not configured - "
                "add a cache_dir parameter to ~/.rekallrc.")

        # Cache dir may be specified relative to the home directory.
        if config.GetHomeDir():
            cache_dir = os.path.join(config.GetHomeDir(), cache_dir)

        if not os.access(cache_dir, os.F_OK | os.R_OK | os.W_OK | os.X_OK):
            try:
                os.makedirs(cache_dir)
            except (IOError, OSError):
                raise io_manager.IOManagerError(
                    "Unable to create or access cache directory %s" % cache_dir)

        # We use an IO manager to manage the cache directory directly.
        self.cache_io_manager = io_manager.DirectoryIOManager(urn=cache_dir,
                                                              session=session)
        self.url_manager = self.DELEGATE(session=session, **kwargs)

        self.CheckUpstreamRepository()

        super(CachingManager, self).__init__(session=session, **kwargs)

    def __str__(self):
        return "Local Cache %s" % self.cache_io_manager

    def CheckInventory(self, name):
        if self.cache_io_manager.CheckInventory(name):
            return True

        return self.url_manager.CheckInventory(name)

    def GetData(self, name, **kwargs):
        if self.cache_io_manager.CheckInventory(name):
            local_age = self.cache_io_manager.Metadata(name).get(
                "LastModified", 0)
            remote_age = self.url_manager.Metadata(name).get("LastModified", 0)

            # Only get the local copy if it is not older than the remote
            # copy. This allows the remote end to update profiles and we will
            # automatically pick the latest.
            if local_age >= remote_age:
                data = self.cache_io_manager.GetData(name)
                # Ensure our local cache looks reasonable.
                if data.get("$METADATA"):
                    return data

        # Fetch the data from our base class and store it in the cache.
        data = self.url_manager.GetData(name, **kwargs)

        # Only store the data in the cache if it looks reasonable. Otherwise we
        # will trash the cache with bad data in case we can not access correct
        # data.
        if data and data.get("$METADATA"):
            self.session.logging.debug("Adding %s to local cache.", name)
            self.cache_io_manager.StoreData(name, data)

        return data

    def StoreData(self, name, data, **options):
        self.cache_io_manager.StoreData(name, data, **options)

    def CheckUpstreamRepository(self):
        """Checks the repository for freshness."""
        upstream_inventory = self.url_manager.inventory

        # This indicates failure to contact the remote repository. In this case
        # we do not want to invalidate our cache, just use the cache as is.
        if not self.url_manager.ValidateInventory():
            self.session.logging.warn(
                "Will attempt to use the local cache. This is likely "
                "to fail if profiles are missing locally!")
            return

        cache_inventory = self.cache_io_manager.inventory
        modified = False

        for item, metadata in cache_inventory.get("$INVENTORY", {}).items():
            upstream_meta = upstream_inventory.get(
                "$INVENTORY", {}).get(item)

            if (upstream_meta is None or
                    upstream_meta["LastModified"] > metadata["LastModified"]):
                cache_inventory["$INVENTORY"].pop(item)
                modified = True

        if modified:
            self.cache_io_manager.FlushInventory()


class CacheDirectoryManager(CachingManager):
    DELEGATE = io_manager.DirectoryIOManager



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# Authors:
# Michael Cohen <scudette@google.com>
#
# Acknowledgments:
# We would like to thank Chakib Gzenayi for his patient testing and suggestions
# in the development of this plugin.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

"""This module implements plugins to inspect Window's DNS resolver cache.

In windows DNS requests are cached by the DNS resolver service. This is a
service running in svchost.exe and implemented as the mostly undocumented DLL
dnsrslvr.dll.

"""
import socket

from rekall import scan
from rekall import utils
from rekall import obj
from rekall.plugins.windows import common

# pylint: disable=protected-access


# Most common DNS types.
DNS_TYPES = {
    1: "A",
    5: "CNAME",
    28: "AAAA",
}

types = {
    "DNS_HASHTABLE_ENTRY": [None, {
        "List": [0x0, ["_LIST_ENTRY"]],
        "Name": [0x8, ["Pointer", dict(
            target="UnicodeString"
            )]],

        "Record": [0x18, ["Pointer", dict(
            target="DNS_RECORD"
        )]],
    }],

    "DNS_RECORD": [None, {
        "Next": [0, ["Pointer", dict(
            target="DNS_RECORD"
            )]],
        "Name": [8, ["Pointer", dict(
            target="UnicodeString"
            )]],
        "Type": [16, ["Enumeration", dict(
            choices=DNS_TYPES,
            target="unsigned short"
        )]],
        "DataLength": [18, ['unsigned short']],
        "Data": [0x20, ['char']],
    }],
}

win10_overlays = {
    "DNS_HASHTABLE_ENTRY": [None, {
        "List": [0x8, ["_LIST_ENTRY"]],
        "Name": [0x38, ["Pointer", dict(
            target="UnicodeString"
            )]],

        "Record": [0x58, ["Pointer", dict(
            target="DNS_RECORD"
        )]],
    }],

}



class DNS_RECORD(obj.Struct):
    @property
    def Data(self):
        if self.Type == "CNAME":
            return self.m("Data").cast(
                "Pointer", target="UnicodeString").deref()
        elif self.Type == "A":
            return utils.inet_ntop(
                socket.AF_INET, self.obj_vm.read(self.m("Data").obj_offset, 4))


def InitializedDNSTypes(profile):
    profile.add_types(types)
    profile.add_types(dict(
        _LIST_ENTRY=profile.session.profile.vtypes["_LIST_ENTRY"]))

    # Windows 10 changes things around a bit.
    if profile.session.profile.metadata("major") == 10:
        profile.add_overlay(win10_overlays)

    profile.add_classes(
        DNS_RECORD=DNS_RECORD
    )

    return profile


class WinDNSCache(common.WindowsCommandPlugin):
    """Dump the windows DNS resolver cache."""

    name = "dns_cache"

    @classmethod
    def is_active(cls, session):
        return (super(WinDNSCache, cls).is_active(session) and
                session.profile.metadata("arch") == "AMD64" and
                session.profile.metadata("major") >= 6)

    @classmethod
    def args(cls, parser):
        super(WinDNSCache, cls).args(parser)
        parser.add_argument("--hashtable", default=None,
                            help="Optionally provide the hashtable")

        parser.add_argument("--index", default=True, type="Boolean",
                            help="Should we use the index")

    def __init__(self, hashtable=None, index=True, **kwargs):
        super(WinDNSCache, self).__init__(**kwargs)
        self.hashtable = hashtable
        self.index = index

    def _find_svchost_vad(self):
        """Returns the vad and _EPROCESS of the dnsrslvr.dll."""
        pslist = self.session.plugins.pslist(proc_regex="svchost.exe")
        for task in pslist.filter_processes():
            self.session.report_progress("Checking pid %s for dnsrslvr.dll",
                                         task.pid)

            for vad in task.RealVadRoot.traverse():
                try:
                    filename = vad.ControlArea.FilePointer.FileName.v()
                    if filename.endswith("\\dnsrslvr.dll"):
                        return vad, task
                except AttributeError:
                    pass

        return None, None

    def _verify_hash_table(self, allocation, heap):
        """Verify the allocation between start and end for a hash table.

        We have observed that often the hash table may contain corrupt data due
        to paging smear during acquisition, hence more rigorous checks might
        actually fail to find the correct hash table due to corrupted data
        confusing the sanity checks here. It is always better to detect the
        correct version using the profile repository.
        """
        self.session.logging.debug("Verifying hash table at %#x",
                                   allocation.obj_offset)

        if self.hashtable and allocation.obj_offset != self.hashtable:
            return False

        # Find all segments in this heap.
        segments = utils.RangedCollection()
        for seg in heap.Segments:
            segments.insert(seg.FirstEntry, seg.LastValidEntry, seg)

        # We usually observe the hash table to be about 1600 bytes, but it might
        # grow.
        if allocation.length > 1600 * 3 or allocation.length < 1600:
            return False

        # Cast the allocation into a hash table.
        cache_hash_table = allocation.cast(
            "Array",
            target="Pointer",
            target_args=dict(
                target="DNS_HASHTABLE_ENTRY",
            ),
            profile=self.profile,
            size=allocation.length)

        count = 0
        for entry in cache_hash_table:
            # If the hashtable entry is null, keep searching.
            entry = entry.v()
            if entry == 0:
                continue

            # ALL entry pointers must point back into one of the other segments
            # in this heap (Since DNS_HASHTABLE_ENTRY are allocated from this
            # heap)..
            dest_segment = segments.get_range(entry)
            if dest_segment is None:
                return False

            count += 1

        # It may be that the hashtable is all empty but otherwise we will match
        # a zero allocated block.
        if count == 0:
            return False

        return cache_hash_table

    def _locate_heap_using_index(self, task):
        """Locate the heap by referencing the index.

        We consult the profile repository for all known versions of dnsrslvr.dll
        and use the known symbol offsets to identify the currently running
        version. Unfortunately often the RSDS section of the PE file is not
        mapped and so we can not directly identify the running version. We
        therefore use the following algorithm:

        1. Enumerate the g_CacheHeap constant for each known version and ensure
        it is referring to a valid heap.

        2. Using the matching profile, dereference the g_HashTable constant and
        ensure it refers to a valid allocation within the above heap.

        3. If both these conditions exist, we return the hash table without
        further checks.

        This method is generally more robust than scanning for the pointers.
        """
        # The base addresses of all valid heaps.
        heaps = set([x.v() for x in task.Peb.ProcessHeaps])
        dnsrslvr_index = self.session.LoadProfile("dnsrslvr/index")

        # The base address of dnsrslvr.dll.
        base_address = self.session.address_resolver.get_address_by_name(
            "dnsrslvr")

        # Now check all profiles for these symbols.
        for profile, symbols in dnsrslvr_index.index.iteritems():
            # Correct symbols offset for dll base address.
            lookup = dict((y[0], x + base_address) for x, y in symbols)

            # According to this profile, where is the cache heap and hash table?
            heap_pointer = self.profile.Pointer(lookup.get("g_CacheHeap")).v()
            hash_tbl_ptr = self.profile.Pointer(lookup.get("g_HashTable")).v()

            if heap_pointer in heaps:
                heap = self.heap_profile._HEAP(
                    offset=heap_pointer
                )

                for entry in heap.Entries:
                    if entry.Allocation.obj_offset == hash_tbl_ptr:
                        self.session.logging.info(
                            "dnsrslvr.dll match profile %s. Hash table is at "
                            "%#x", profile, hash_tbl_ptr)

                        return self.profile.Array(
                            hash_tbl_ptr,
                            target="Pointer",
                            target_args=dict(
                                target="DNS_HASHTABLE_ENTRY",
                            ),
                            count=entry.Allocation.length / 8)

        self.session.logging.info(
            "Failed to detect the exact version of dnsrslvr.dll, please "
            "consider sending a copy of this DLL's GUID to the Rekall team so "
            "we can add it to the index.")

    def _locate_heap(self, task, vad):
        """Locate the correct heap by scanning for its reference.

        Find the references into the heap from the dnsrslvr.dll vad. This will
        normally be stored in dnsrslvr.dll's global variable called g_CacheHeap.
        """
        scanner = scan.PointerScanner(
            pointers=task.Peb.ProcessHeaps,
            session=self.session,
            address_space=self.session.GetParameter("default_address_space"))

        seen = set()
        for hit in scanner.scan(vad.Start, maxlen=vad.Length):
            heap = self.heap_profile.Pointer(
                hit, target="_HEAP"
            ).deref()


            if heap in seen:
                continue

            seen.add(heap)

            for entry in heap.Entries:
                hash_table = self._verify_hash_table(entry.Allocation, heap)
                if hash_table:
                    return hash_table

    def locate_cache_hashtable(self):
        """Finds the DNS cache hashtable.

        The dns cache runs inside one of the svchost.exe processes and is
        implemented via the dnsrslvr.dll service. We therefore first search for
        the correct VAD region for this DLL. We then find the private heap that
        belongs to the resolver.
        """
        vad, task = self._find_svchost_vad()
        if task is None:
            raise RuntimeError("Unable to find svchost.exe for dnsrslvr.dll.")

        # Switch to the svchost process context now.
        self.cc.SwitchProcessContext(task)

        # Load the profile for dnsrslvr and add the new types.
        self.profile = InitializedDNSTypes(
            self.session.address_resolver.LoadProfileForName("dnsrslvr"))

        hash_table = self.session.address_resolver.get_constant_object(
            "dnsrslvr!g_HashTable",
            "Pointer",
            target_args=dict(
                target="Array",
                target_args=dict(
                    count=self.session.address_resolver.get_constant_object(
                        "dnsrslvr!g_HashTableSize", "unsigned int").v(),
                    target="Pointer",
                    target_args=dict(
                        target="DNS_HASHTABLE_ENTRY"
                    )
                )
            )
        )
        if hash_table:
            return hash_table.deref()

        self.heap_profile = self.session.address_resolver.LoadProfileForName(
            "ntdll")

        # First try to locate the hash table using the index, then fallback to
        # using scanning techniques:
        if self.index:
            hash_table = self._locate_heap_using_index(task)
            if hash_table:
                return hash_table

        return self._locate_heap(task, vad)

    def render(self, renderer):
        self.cc = self.session.plugins.cc()
        with self.cc:
            cache_hash_table = self.locate_cache_hashtable()
            if cache_hash_table:
                renderer.table_header([
                    dict(name="Name", type="TreeNode", width=45),
                    ("Record", "record", "[addrpad]"),
                    ("Type", "type", "16"),
                    ("Data", "data", ""),
                ])

                for bucket in cache_hash_table:
                    for entry in bucket.List.list_of_type_fast(
                            "DNS_HASHTABLE_ENTRY", "List",
                            include_current=True):
                        name = entry.Name.deref()
                        renderer.table_row(
                            name, entry, "HTABLE", depth=0)
                        for record in entry.Record.walk_list("Next", True):
                            name = record.Name.deref() or name
                            renderer.table_row(
                                name,
                                record,
                                record.Type,
                                record.Data, depth=1)



#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
Helper functions to make defining components nicer.

Exists solely to support rekall.entities.definitions.
"""

__author__ = "Adam Sindelar <adamsh@google.com>"

import datetime

from rekall import obj
from rekall import registry


class TypeDescriptor(object):
    """Defines a type descriptor, which can coerce values into target type."""

    type_name = None
    __metaclass__ = registry.MetaclassRegistry

    def chill_coerce(self, value):
        """Like coerce, but is chill about getting exceptions.

        Will returned coerced value if it can, otherwise will return value.
        """
        try:
            return self.coerce(value)
        except TypeError:
            return value

    def coerce(self, value):
        """Return value as this type or raise TypeError if not convertible."""
        return value

    def sortkey(self, coerced):
        return coerced

    def __repr__(self):
        return "%s" % type(self).__name__

    def __unicode__(self):
        return repr(self)

    def __str__(self):
        return repr(self)

    def __eq__(self, other):
        return self.type_name == other.type_name

    def __ne__(self, other):
        return not self.__eq__(other)


class ScalarDescriptor(TypeDescriptor):
    """Take an instance of type and calls its constructor to coerce."""

    default_instance = None
    type_cls = int

    def __init__(self, type_cls=None):
        super(ScalarDescriptor, self).__init__()
        if type_cls:
            self.type_cls = type_cls
        self.type_name = self.type_cls.__name__

    def coerce(self, value):
        if value == None:
            return None

        if isinstance(value, self.type_cls):
            return value

        return self.type_cls(value)

    def sortkey(self, coerced):
        if not coerced:
            return self.default_instance

        return coerced

    def __repr__(self):
        return "%s (scalar type)" % self.type_cls.__name__


class UnicodeDescriptor(TypeDescriptor):
    type_cls = unicode
    type_name = "unicode"

    def coerce(self, value):
        if value == None:
            return None

        # Already in unicode.
        elif isinstance(value, unicode):
            return value

        elif isinstance(value, str):
            return unicode(value, "utf-8", "ignore")

        # Call the object's __unicode__ method if needed.
        return unicode(value)


class DatetimeDescriptor(ScalarDescriptor):
    """Python's datetime is a special snowflake that requires some massaging.

    This is just like ScalarDescriptor, except it works around the weirdness
    in datetime's sorting behavior and absence of default parameters in the
    constructor.
    """

    default_instance = datetime.datetime.fromtimestamp(0)
    type_cls = datetime.datetime

    def sortkey(self, coerced):
        return super(DatetimeDescriptor, self).sortkey(coerced).strftime("%s")


class BaseObjectDescriptor(TypeDescriptor):
    """Makes sure base objects are dereferenced."""

    def __init__(self):
        super(BaseObjectDescriptor, self).__init__()
        self.type_cls = obj.BaseObject
        self.type_name = "BaseObject"

    def coerce(self, value):
        if not value:
            return None

        superposition_typedesc = getattr(value, "typedesc", None)
        if superposition_typedesc and isinstance(superposition_typedesc,
                                                 type(self)):
            # TODO: This is kind of incorrect, in that it doesn't actually
            # coerce the variants. However, it is unlikely to cause any
            # problems at present time, and this entire module is going to be
            # largely rewritten using EFILTER types soon, which will address
            # the handling of superpositions in a much cleaner manner.
            return value

        if not isinstance(value, obj.BaseObject):
            raise TypeError("%s is not a BaseObject." % (value, ))

        if isinstance(value, obj.Pointer):
            return value.deref()

        return value

    def __repr__(self):
        return "BaseObject type"


class PointerDescriptor(TypeDescriptor):
    """Stores a pointer."""

    def __init__(self):
        super(PointerDescriptor, self).__init__()
        self.type_cls = obj.Pointer
        self.type_name = "Pointer"

    def coerce(self, value):
        if not value:
            return value

        if isinstance(value, obj.Pointer):
            return value

        if isinstance(value, obj.BaseObject):
            return value.obj_profile.Pointer(value=value.obj_offset,
                                             vm=value.obj_vm,
                                             target=value.obj_type)

        raise TypeError("%s is not a Pointer to a BaseObject." % value)


class NoneDescriptor(TypeDescriptor):
    "NoneDescriptor doesn't care."

    def coerce(self, value):
        return value

    def __repr__(self):
        return "untyped (NoneDescriptor)"


class TupleDescriptor(TypeDescriptor):
    """Declared for tuple types; coerces each member to its respective type."""

    type_name = "tuple"

    def __init__(self, tpl):
        super(TupleDescriptor, self).__init__()
        self.types = [TypeFactory(x) for x in tpl]

    def coerce(self, value):
        return tuple(self.types[i].coerce(x) for i, x in enumerate(value))

    def __repr__(self):
        return "(%s)" % ", ".join(self.types)


class ListDescriptor(TypeDescriptor):
    """Declared for nested types (e.g. list of ints)."""

    type_name = "list"

    def __init__(self, member_type):
        super(ListDescriptor, self).__init__()
        self.member_type = TypeFactory(member_type)

    def coerce(self, value):
        if value is None:
            return frozenset()
        return frozenset([self.member_type.coerce(x) for x in value])

    def __repr__(self):
        return "[%s] (list type)" % self.member_type


class EnumDescriptor(TypeDescriptor):
    """Defines an enum type for a component attribute."""

    type_name = "str"

    def __init__(self, *args):
        super(EnumDescriptor, self).__init__()
        self.legal_values = args

    def coerce(self, value):
        if value == None:
            return value

        if value not in self.legal_values:
            raise TypeError(
                "%s is not a valid value for enum %s" % (value,
                                                         self.legal_values))

        return value

    def __repr__(self):
        return "{%s} (enum type)" % ", ".join(self.legal_values)


def TypeFactory(type_desc):
    """Creates the appropriate TypeDescriptor or subclass instance.

    If given a type instance, will create TypeDescriptor (most common use).

    If given a set, will interpret it as enum and create EnumDescriptor.

    If given a tuple, will interpret as composite attribute and create a
    TupleDescriptor.

    If given a list, will interpret is as a nested type and create a
    ListDescriptor.
    """
    if isinstance(type_desc, TypeDescriptor):
        # Fall through for stuff defined explictly.
        return type_desc

    if type_desc == unicode:
        return UnicodeDescriptor()

    if isinstance(type_desc, type):
        return ScalarDescriptor(type_desc)

    if isinstance(type_desc, str):
        return TypeDescriptor.classes[type_desc]()

    if type_desc is None:
        return NoneDescriptor()

    if isinstance(type_desc, set):
        return EnumDescriptor(*type_desc)

    if isinstance(type_desc, tuple):
        return TupleDescriptor(type_desc)

    if isinstance(type_desc, list):
        return ListDescriptor(type_desc[0])

    raise TypeError("%s is not a valid type descriptor.", type_desc)



# Rekall Memory Forensics
#
# Copyright 2014 Google Inc. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
The Rekall Entity Layer.
"""
__author__ = "Adam Sindelar <adamsh@google.com>"

from rekall.entities import entity

from efilter import errors
from efilter import engine


class ValidationError(errors.EfilterError):
    pass


class QueryValidator(engine.VisitorEngine):
    """Checks the query for semantic errors.

    This checks for the following:

    - Attribute/literal type mismatch.
    - Non-existent components or attributes.
    """

    def error(self, message, exp):
        raise ValidationError(error=message, start=exp.start, end=exp.end,
                              query=self.query.source)

    def visit_Expression(self, exp):
        _ = exp
        return True

    def visit_Binding(self, exp):
        if exp.value.startswith("&"):
            field = entity.Entity.reflect_attribute(exp.value[1:])
            if field and field.typedesc.type_name != "Entity":
                return self.error(
                    "%s is type %s. Reverse lookups require type Entity." %
                    (field.name, field.typedesc.type_name),
                    exp)
        else:
            field = entity.Entity.reflect_attribute(exp.value)

        if field is None:
            return self.error("Attribute %s doesn't exist." % exp.value, exp)

        return True

    def visit_ComponentLiteral(self, exp):
        component = entity.Entity.reflect_component(exp.value)
        if component is None:
            return self.error("Component %s doesn't exist." % exp.value, exp)

        return True

engine.Engine.register_engine(QueryValidator, "validator")



"""Entity collectors."""

from rekall.plugins.collectors import ballast
from rekall.plugins.collectors import darwin
from rekall.plugins.collectors import events
from rekall.plugins.collectors import windows
